# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from .acs import ACS
from .fetch import fetch_first, fetch_last

__all__ = ["ACS", "fetch_first", "fetch_last"]



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/active_contracts_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import event_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_event__pb2
from . import transaction_filter_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_transaction__filter__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n5com/daml/ledger/api/v1/active_contracts_service.proto\x12\x16\x63om.daml.ledger.api.v1\x1a\"com/daml/ledger/api/v1/event.proto\x1a/com/daml/ledger/api/v1/transaction_filter.proto\"\x95\x01\n\x19GetActiveContractsRequest\x12\x1b\n\tledger_id\x18\x01 \x01(\tR\x08ledgerId\x12\x41\n\x06\x66ilter\x18\x02 \x01(\x0b\x32).com.daml.ledger.api.v1.TransactionFilterR\x06\x66ilter\x12\x18\n\x07verbose\x18\x03 \x01(\x08R\x07verbose\"\xa6\x01\n\x1aGetActiveContractsResponse\x12\x16\n\x06offset\x18\x01 \x01(\tR\x06offset\x12\x1f\n\x0bworkflow_id\x18\x02 \x01(\tR\nworkflowId\x12O\n\x10\x61\x63tive_contracts\x18\x03 \x03(\x0b\x32$.com.daml.ledger.api.v1.CreatedEventR\x0f\x61\x63tiveContracts2\x97\x01\n\x16\x41\x63tiveContractsService\x12}\n\x12GetActiveContracts\x12\x31.com.daml.ledger.api.v1.GetActiveContractsRequest\x1a\x32.com.daml.ledger.api.v1.GetActiveContractsResponse0\x01\x42\x9a\x01\n\x16\x63om.daml.ledger.api.v1B ActiveContractsServiceOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3')



_GETACTIVECONTRACTSREQUEST = DESCRIPTOR.message_types_by_name['GetActiveContractsRequest']
_GETACTIVECONTRACTSRESPONSE = DESCRIPTOR.message_types_by_name['GetActiveContractsResponse']
GetActiveContractsRequest = _reflection.GeneratedProtocolMessageType('GetActiveContractsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETACTIVECONTRACTSREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.active_contracts_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.GetActiveContractsRequest)
  })
_sym_db.RegisterMessage(GetActiveContractsRequest)

GetActiveContractsResponse = _reflection.GeneratedProtocolMessageType('GetActiveContractsResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETACTIVECONTRACTSRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.active_contracts_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.GetActiveContractsResponse)
  })
_sym_db.RegisterMessage(GetActiveContractsResponse)

_ACTIVECONTRACTSSERVICE = DESCRIPTOR.services_by_name['ActiveContractsService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.daml.ledger.api.v1B ActiveContractsServiceOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\252\002\026Com.Daml.Ledger.Api.V1'
  _GETACTIVECONTRACTSREQUEST._serialized_start=167
  _GETACTIVECONTRACTSREQUEST._serialized_end=316
  _GETACTIVECONTRACTSRESPONSE._serialized_start=319
  _GETACTIVECONTRACTSRESPONSE._serialized_end=485
  _ACTIVECONTRACTSSERVICE._serialized_start=488
  _ACTIVECONTRACTSSERVICE._serialized_end=639
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from asyncio import Future, InvalidStateError, ensure_future, gather, get_event_loop
from collections import defaultdict
from dataclasses import dataclass, field, replace
from datetime import timedelta
from enum import Enum, auto
from functools import wraps
import inspect
import operator
from threading import RLock
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Awaitable,
    Callable,
    Collection,
    DefaultDict,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
    overload,
)
from uuid import uuid4
import warnings

from .. import LOG
from ..ledger import Command, CreateEvent, ExerciseResponse
from ..prim import Party
from ..protocols.events import BaseEvent
from ..util.asyncio_util import LongRunningAwaitable, Signal, completed, failed, propagate
from .commands import CommandBuilder
from .events import EventKey

if TYPE_CHECKING:
    from .api import PartyClient

DEFAULT_BOT_STOP_TIMEOUT = timedelta(seconds=30)


E = TypeVar("E", bound=BaseEvent)
BotCallback = Callable[[E], Any]
BotFilter = Callable[[E], bool]


@dataclass(frozen=True)
class SourceLocation:
    file_name: Optional[str]
    start_line: int
    end_line: int


class _BotRunLevel(Enum):
    CONTINUE = auto()
    SUSPEND = auto()
    TERMINATE = auto()


class Bot:
    def __init__(self, party_client: "Optional[PartyClient]", name: str):
        self._handlers: DefaultDict[str, List[BotEntry]] = defaultdict(list)
        self._party_client = party_client
        self._id = uuid4().hex
        self._name = name
        self._queue = []  # type: List[BotInvocation]
        self._signal = None  # type: Optional[Signal]
        self._idle = True
        self._run_level = _BotRunLevel.CONTINUE

    def event_keys(self) -> "AbstractSet[str]":
        """
        Return the set of keys that event handlers in this bot are configured to handle.
        """
        return frozenset(self._handlers)

    def wants_any_keys(self, keys: "Collection[str]") -> bool:
        return bool(set(keys).intersection(self._handlers))

    def add_event_handler(
        self,
        keys: "Union[str, Collection[str]]",
        handler: "BotCallback",
        filter_fn: "Optional[BotFilter]" = None,
    ) -> None:
        """
        Add a new event handler to this bot for the specified event.

        :param keys:
            The key(s) of the event (as returned by :meth:`EventKey.from_event`).
        :param filter_fn:
            An optional callback that returns `True` or `False` on whether the corresponding
            callback should be invoked. This cannot be a coroutine function.
        :param handler:
            An event handler to be invoked when an event with the specified key is raised.
        """
        if isinstance(keys, str):
            keys = [keys]
        else:
            keys = tuple(keys)
            for key in keys:
                if not isinstance(key, str):
                    raise ValueError("expected a string key")
        if not callable(handler):
            raise ValueError("handler must be callable")

        source_loc = None  # type: Optional[SourceLocation]
        # noinspection PyBroadException
        try:
            source_file = inspect.getsourcefile(inspect.unwrap(handler))
            lines, start_line = inspect.getsourcelines(handler)
            end_line = start_line + len(lines)
            source_loc = SourceLocation(source_file, start_line, end_line)
        except Exception:  # noqa
            LOG.warning(
                "Could not compute original source information for %r", handler, exc_info=True
            )

        if self._party_client is not None:
            # noinspection PyProtectedMember
            impl = self._party_client._impl
            handler = wrap_as_command_submission(impl.write_commands, handler, filter_fn)

        for key in keys:
            self._handlers[key].append(BotEntry(key, handler, filter_fn, source_loc))

    def ledger_created(self, template: Any):
        def _register_created(fn):
            self.add_event_handler(EventKey.contract_created(True, template), fn)

        return _register_created

    async def _main(self):
        """
        The "main" coroutine of the bot. Invokes handlers on each event as they come in.
        """
        # noinspection PyBroadException
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", DeprecationWarning)
                # noinspection PyDeprecation
                self._signal = Signal()
            while self._run_level != _BotRunLevel.TERMINATE:
                # the main queue contains either events we have not yet processed yet or ``None``
                # markers that merely indicate running status should be "re-checked"
                self._idle = False
                while self._queue:
                    invocation = self._queue.pop(0)
                    # noinspection PyBroadException
                    try:
                        fut = ensure_future(self._handle_event(invocation.event))
                        propagate(fut, invocation.future)
                        await invocation.future
                    except Exception:  # noqa
                        LOG.exception(
                            f"An event handler in a bot has thrown an exception! "
                            f"(offending event: {invocation.event})"
                        )
                self._idle = True
                await self._signal.wait()
            self._signal = None
        except Exception:  # noqa
            LOG.exception("A bot thread died abnormally.")

    async def _handle_event(self, event: "BaseEvent") -> None:
        """
        Process an event, mostly by calling appropriate callbacks.

        :param event: The event to process.
        """
        # The _PartyClientImpl and Bot classes are the ones raising events, but to the user, they
        # registered from the perspective of a client with a certain thread affinity. Replace the
        # "source" of the event with the client that the user originally used.
        if self._party_client is not None:
            new_event = replace(event, client=self._party_client)
        else:
            new_event = event

        for event_key in EventKey.from_event(new_event):
            LOG.debug("Party %s dispatching event %r to its bots...", self.party, event_key)
            handlers = self._handlers.get(event_key)
            if handlers is not None:
                for handler in handlers:
                    # noinspection PyBroadException
                    try:
                        if handler.filter is None or handler.filter(new_event):
                            fut = handler.callback(new_event)
                            if inspect.isawaitable(fut):
                                await fut
                    except Exception:  # noqa
                        LOG.exception("An event handler in a bot has thrown an exception!")
        LOG.debug("Party %s finished handling events.", self.party)

    def notify(self, event: "BaseEvent") -> "Awaitable[None]":
        """
        Notifies handler(s) associated with this bot that the given event has occurred. Note that
        this notification is asynchronous: in other words, event handlers will not have processed
        this event by the time this function returns.

        :param event: The event to raise.
        """
        if not isinstance(event, BaseEvent):
            raise ValueError("expected a BaseEvent")

        if self._queue is None:
            raise InvalidStateError("Cannot notify on a bot whose main method is not running")
        bot_invocation = BotInvocation(event)
        self._queue.append(bot_invocation)
        if self._signal is not None:
            self._signal.notify_all()
        return bot_invocation.future

    def _dispatch(self, event: "BaseEvent") -> None:
        for event_key in EventKey.from_event(event):
            self._handlers.get(event_key)

    # <editor-fold desc="State control functions">

    def pause(self) -> None:
        """
        Immediately change the state of this bot to ``PAUSING``, and pause event handler
        invocations. The event handler currently running is allowed to complete. When that is
        completed, the state is changed to ``PAUSED``.
        """
        LOG.info("Pausing the bot thread for party %r...", self.party)
        self._run_level = _BotRunLevel.SUSPEND
        if self._signal is not None:
            self._signal.notify_all()

    def resume(self) -> None:
        """
        Immediately change the state of this bot to ``RESUMING`` and process any events that have
        queued up while the bot was paused. When this queue is fully drained, the state is changed
        to ``RUNNING``.
        """
        LOG.info("Resuming the bot thread for party %r...", self.party)
        self._run_level = _BotRunLevel.CONTINUE
        if self._signal is not None:
            self._signal.notify_all()

    def stop(self):
        """
        Permanently stop this bot. If you need to be able to "restart" a stopped bot, use
        :meth:`pause` and :meth:`resume` instead.
        """
        LOG.info("Stopping the bot thread for party %r...", self.party)
        self._run_level = _BotRunLevel.TERMINATE
        if self._signal is not None:
            self._signal.notify_all()

    # </editor-fold>

    # <editor-fold desc="State query properties/functions">

    @property
    def id(self) -> str:
        """
        The ID of this bot, generated at runtime.
        """
        return self._id

    @property
    def name(self):
        """
        The name of this bot. Defaults to the name of the original event handler if unspecified.
        """
        return self._name

    @property
    def party(self) -> "Optional[Party]":
        """
        Primary party that this bot receives events for (and potentially generates commands for).
        """
        return self._party_client.party if self._party_client is not None else None

    def entries(self) -> "Sequence[BotEntry]":
        """
        The collection of individual event handlers in a bot, in the order that they will be
        executed.
        """
        return tuple(entry for collection in self._handlers.values() for entry in collection)

    @property
    def state(self) -> "BotState":
        """
        Current running state of the bot.
        """
        if self._run_level == _BotRunLevel.CONTINUE:
            return BotState.RUNNING if self._signal is not None else BotState.STARTING
        elif self._run_level == _BotRunLevel.TERMINATE:
            return BotState.STOPPED if self._idle else BotState.STOPPING
        elif self._run_level == _BotRunLevel.SUSPEND:
            return BotState.PAUSED if self._idle else BotState.PAUSING
        else:
            raise RuntimeError("unexpected _BotRunLevel enum value")

    @property
    def running(self) -> bool:
        """
        Return ``True`` if this bot is currently processing events.
        """
        return self._signal is not None and (
            self._run_level == _BotRunLevel.CONTINUE or not self._idle
        )

    # </editor-fold>


class BotCollection(Sequence[Bot]):
    """
    A collection of bots for a party.

    This class is thread-safe except for :meth:`notify` and :meth:`_main` in order to support adding
    event handlers from any thread. The most common use of this is for ``SimplePartyClient``
    instances, where event registration is done from the main thread (from the perspective of the
    caller) and event notifications are done on an asyncio event loop thread (hidden from the
    caller).
    """

    def __init__(self, party: "Optional[Party]"):
        self.party = party
        self._bots = []  # type: List[Bot]
        self._fut = None  # type: Optional[LongRunningAwaitable]
        self._lock = RLock()

    def __len__(self) -> int:
        with self._lock:
            return len(self._bots)

    @overload
    def __getitem__(self, i: int) -> Bot:
        ...

    @overload
    def __getitem__(self, i: slice) -> Sequence[Bot]:
        ...

    def __getitem__(self, i):
        with self._lock:
            return operator.getitem(self._bots, i)

    def __iter__(self):
        with self._lock:
            bots = list(self._bots)
        return iter(bots)

    def add_new(self, name: str, party_client: "Optional[PartyClient]" = None) -> "Bot":
        bot = Bot(party_client, name)
        with self._lock:
            self._bots.append(bot)
            if self._fut is not None:
                # the _main coroutine has already started, so just add this bot's main method to the
                # set of coroutines we track
                # noinspection PyProtectedMember
                self._fut.append(ensure_future(bot._main()))
        return bot

    def add_single(
        self,
        keys: "Union[str, Sequence[str]]",
        handler: "BotCallback",
        filter_fn: "Optional[BotFilter]" = None,
        name: "Optional[str]" = None,
        party_client: "Optional[PartyClient]" = None,
    ) -> "Bot":
        """
        Convenience method for creating a bot with a single event handler.
        """
        if name is None:
            # by default, the name of a single-event-handler bot is simply the name of the passed
            # in function
            name = handler.__name__
        if isinstance(keys, str):
            keys = [keys]

        bot = self.add_new(name, party_client)
        for key in keys:
            bot.add_event_handler(key, handler, filter_fn)
        return bot

    def notify(self, event: "BaseEvent"):
        futures = []
        try:
            event_keys = EventKey.from_event(event)
            for bot in self._bots:
                if bot.wants_any_keys(event_keys):
                    futures.append(ensure_future(bot.notify(event)))
        except:  # noqa
            # This exception indicates a problem with the scheduling code and not user bot code.
            # Exceptions thrown by user code would be propagated through the Future
            LOG.exception("Failed to notify event for party %s: %r", self.party, event)
            raise

        if len(futures) == 0:
            return completed(None)
        elif len(futures) == 1:
            return futures[0]
        else:
            return gather(*futures, return_exceptions=True)

    async def _main(self):
        if self.party is not None:
            LOG.info("Party %s bots coroutine started.", self.party)
        else:
            LOG.info("Network bots coroutine started.")

        with self._lock:
            self._fut = LongRunningAwaitable()
            # noinspection PyProtectedMember
            self._fut.extend(ensure_future(bot._main()) for bot in self._bots)

        await self._fut

        with self._lock:
            self._fut = None

        if self.party is not None:
            LOG.info("Party %s bots coroutine finished.", self.party)
        else:
            LOG.info("Network bots coroutine finished.")

    def stop_all(self):
        with self._lock:
            # LongRunningAwaitable suspends itself until all futures are completed. It also keeps
            # itself open waiting for the very first future. In the case that no other futures have
            # been added, take this opportunity to feed LongRunningAwaitable a future that
            # immediately resolves, which will have the effect of resolving that overall future
            # if there are no other futures (or the other futures are resolved as well); otherwise
            # this has no effect.
            self._fut.append(completed(None))
            for bot in list(self._bots):
                bot.stop()
            self._bots.clear()


@dataclass(frozen=True)
class BotEntry:
    event_key: str
    callback: BotCallback
    filter: Optional[BotFilter] = None
    source_location: Optional[SourceLocation] = None


@dataclass(frozen=True)
class BotInvocation:
    event: BaseEvent
    future: Future = field(default_factory=lambda: get_event_loop().create_future())


class BotState(Enum):
    """
    Possible states of a :class:`Bot`.
    """

    STARTING = "STARTING"  #: The bot is starting (has not yet received the "ready" event).
    PAUSING = "PAUSING"  #: This bot has been told to pause, but has not yet completed processing events in flight.
    PAUSED = "PAUSED"  #:
    RESUMING = "RESUMING"
    RUNNING = "RUNNING"
    STOPPING = "STOPPING"
    STOPPED = "STOPPED"


# noinspection PyShadowingBuiltins
def wrap_as_command_submission(
    submit_fn, callback, filter
) -> Callable[[BaseEvent], Awaitable[Any]]:
    """
    Normalize a callback to something that takes a single contract ID and contract data, and
    return an awaitable that is resolved when the underlying command has been fully submitted.
    """
    import inspect

    @wraps(callback)
    def implementation(*args, **kwargs):
        if filter is not None and not filter(*args, **kwargs):
            return completed(None)

        try:
            ret = callback(*args, **kwargs)
        except BaseException as exception:
            LOG.exception("The callback %r threw an exception!", callback)
            return failed(exception)

        if ret is None:
            return completed(None)
        elif isinstance(ret, (CommandBuilder, Command, list, tuple)):
            try:
                ret_fut = submit_fn(ret)
            except BaseException as exception:
                LOG.exception(
                    "The callback %r returned commands that could not be submitted! (%s)",
                    callback,
                    ret,
                )
                return failed(exception)
            return ret_fut
        elif inspect.isawaitable(ret):
            # the user-provided callback returned an Awaitable
            cmd_fut = ensure_future(ret)
            if cmd_fut.done():
                if cmd_fut.cancelled() or cmd_fut.exception() is not None:
                    # a cancelled or failed user-provided callback Future is the same as the
                    # command submission itself failing
                    return cmd_fut

                # functionally equivalent to the non-Awaitable case if the Awaitable has already
                # completed
                return submit_fn(cmd_fut.result())
            else:
                # create `fut`, which we'll give to the user; wait for `cmd_fut` to finish, then
                # take the result of that awaitable and try to submit a command with that result
                fut = get_event_loop().create_future()

                def cmd_future_finished(_):
                    ret = cmd_fut.result()
                    if ret is None:
                        fut.set_result(None)
                    elif isinstance(ret, (CreateEvent, ExerciseResponse)):
                        fut.set_result(ret)
                    elif isinstance(ret, (CommandBuilder, Command, list, tuple)):
                        propagate(ensure_future(submit_fn(ret)), fut)
                    elif inspect.isawaitable(ret):
                        LOG.error("A callback cannot return an Awaitable of an Awaitable")
                        raise InvalidStateError(
                            "A callback cannot return an Awaitable of an Awaitable"
                        )

                cmd_fut.add_done_callback(cmd_future_finished)

                return fut
        else:
            LOG.error("the callback %r returned a value of an unexpected type: %s", callback, ret)
            raise ValueError("unexpected return type from a callback")

    return implementation



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer, RepeatedScalarFieldContainer
from google.protobuf.message import Message as _Message
from google.protobuf.timestamp_pb2 import Timestamp

from .completion_pb2 import Completion
from .ledger_offset_pb2 import LedgerOffset

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "CompletionStreamRequest",
    "CompletionStreamResponse",
    "Checkpoint",
    "CompletionEndRequest",
    "CompletionEndResponse",
]


class CompletionStreamRequest(_Message):
    ledger_id: _builtins.str
    application_id: _builtins.str
    @property
    def parties(self) -> RepeatedScalarFieldContainer[_builtins.str]: ...
    @property
    def offset(self) -> LedgerOffset: ...
    def __init__(self, *, ledger_id: _typing.Optional[_builtins.str] = ..., application_id: _typing.Optional[_builtins.str] = ..., parties: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., offset: _typing.Optional[LedgerOffset] = ...): ...
    def HasField(self, field_name: _L["ledger_id", "application_id", "parties", "offset"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["ledger_id", "application_id", "parties", "offset"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class CompletionStreamResponse(_Message):
    @property
    def checkpoint(self) -> Checkpoint: ...
    @property
    def completions(self) -> RepeatedCompositeFieldContainer[Completion]: ...
    def __init__(self, *, checkpoint: _typing.Optional[Checkpoint] = ..., completions: _typing.Optional[_typing.Iterable[Completion]] = ...): ...
    def HasField(self, field_name: _L["checkpoint", "completions"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["checkpoint", "completions"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class Checkpoint(_Message):
    @property
    def record_time(self) -> Timestamp: ...
    @property
    def offset(self) -> LedgerOffset: ...
    def __init__(self, *, record_time: _typing.Optional[Timestamp] = ..., offset: _typing.Optional[LedgerOffset] = ...): ...
    def HasField(self, field_name: _L["record_time", "offset"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["record_time", "offset"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class CompletionEndRequest(_Message):
    ledger_id: _builtins.str
    def __init__(self, *, ledger_id: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["ledger_id"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["ledger_id"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class CompletionEndResponse(_Message):
    @property
    def offset(self) -> LedgerOffset: ...
    def __init__(self, *, offset: _typing.Optional[LedgerOffset] = ...): ...
    def HasField(self, field_name: _L["offset"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["offset"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from io import BytesIO
from os import PathLike
from pathlib import Path
import threading
from typing import AbstractSet, BinaryIO, Collection, Generator, Mapping, Optional, Union
import warnings
from zipfile import ZipFile

from .._gen.com.daml.daml_lf_1_14 import daml_lf_pb2 as pb
from .daml_lf_1 import Archive, Package, PackageRef
from .parse import parse_archive

# Wherever the API expects a DAR, we can take a file path, `bytes`, or a byte buffer.
Dar = Union[bytes, str, Path, BinaryIO]

__all__ = ["Dar", "DarFile", "CachedDarFile", "get_dar_package_ids"]


class DarFile:
    """
    Provides access to the contents of a .dar file.

    This conforms to the :class:`PackageProvider` protocol.

    This class is _not_ thread-safe.
    """

    filename: Optional[str]

    def __init__(self, dar: "Dar"):
        """
        Initialize a new DarFile.

        :param dar:
            A path to a file (either expressed as str or pathlib.Path), a bytes blob, or a buffer.
        """
        if isinstance(dar, str):
            self.filename = dar
            self._buf = None
            self._zip = ZipFile(self.filename)

        elif isinstance(dar, PathLike):
            self.filename = dar.__fspath__()
            self._buf = None
            self._zip = ZipFile(self.filename)

        elif isinstance(dar, bytes):
            self.filename = None
            self._buf = BytesIO(dar)
            self._zip = ZipFile(self._buf)

        elif hasattr(dar, "read"):
            # file-like object (buffers, files, etc.)
            self.filename = None
            self._buf = None
            self._zip = ZipFile(dar)

        else:
            raise TypeError("DarFile only understands file paths or binary blobs")

    def __enter__(self) -> "DarFile":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self._zip.close()
        if self._buf is not None:
            self._buf.close()

    def manifest(self) -> "Optional[Mapping[str, str]]":
        """
        Return the contents of the manifest of this DAR.
        """
        names = self._zip.namelist()
        if "META-INF/MANIFEST.MF" in names:
            manifest_bytes = self._zip.read("META-INF/MANIFEST.MF")
            manifest = {}
            for line in manifest_bytes.decode("utf-8").splitlines():
                name, _, value = line.partition(":")
                manifest[name] = value.strip()
            return manifest
        else:
            return None

    def sdk_version(self) -> "Optional[str]":
        """
        Return the SDK version used to compile this dar (if this information is available).
        """
        manifest = self.manifest()
        return manifest.get("Sdk-Version") if manifest is not None else None

    def archives(self) -> "Collection[Archive]":
        """
        Return :class:`Archive` instances from this :class:`DarFile`.

        :class:`CachedDarFile` is a better choice than `DarFile` if this method is frequently called
        on the same :class:`DarFile`, as package parsing is an expensive operation.
        """
        return [parse_archive(PackageRef(a.hash), a.payload) for a in self._pb_archives()]

    def package(self, package_id: "PackageRef") -> "Package":
        """
        Return the :class:`Package` corresponding to the specified :class:`PackageRef`.

        This function is somewhat inefficient as it must check every DALF for the specified package
        ID. The :method:`DarFile.archives` function should be used if it is known that all archives
        in a DAR are to be accessed, or use a :class:`CachedDarFile` instance if random access of
        packages by package ID are needed.
        """
        payload = self.package_bytes(package_id)
        return parse_archive(package_id, payload).package

    def package_bytes(self, package_id: "PackageRef") -> bytes:
        """
        Return bytes corresponding to the specified :class:`PackageRef`. If this DAR were to be
        uploaded to a ledger, these are the bytes that would be returned for the specified
        :class:`PackageRef`.

        Note that this NOT the same as simply returning a ``.dalf`` file within a DAR, because a
        ``.dalf`` contains an envelope in the form of ``ArchivePayload`` messages, and it is the
        _contents_ of this message that are persisted by Ledger API implementations.
        """
        for a in self._pb_archives():
            if a.hash == package_id:
                return a.payload

        # We do not raise PackageNotFoundError here (even though it seems like it would be a more
        # apt error) because PackageNotFoundError implies the operation is retryable
        raise Exception(f"package not found in a DAR: {package_id!r}")

    def package_ids(self) -> "AbstractSet[PackageRef]":
        """
        Return the set of package IDs from this DAR.
        """
        return frozenset(PackageRef(a.hash) for a in self._pb_archives())

    def async_package_service(self) -> "DarFileAsyncPackageService":
        """
        Return an implementation of :class:`dazl.ledger.aio.PackageService` that is backed by
        the contents of this :class:`DarFile`.
        """
        return DarFileAsyncPackageService(self)

    def blocking_package_service(self) -> "DarFileBlockingPackageService":
        """
        Return an implementation of :class:`dazl.ledger.blocking.PackageService` that is backed by
        the contents of this :class:`DarFile`.

        The returned service is thread-safe as long as this :class:`DarFile` is not accessed
        concurrently by other processes.
        """
        return DarFileBlockingPackageService(self)

    async def get_package(self, package_id: "PackageRef") -> bytes:
        """
        This method is deprecated; use :meth:`DarFile.async_package_service` instead.
        """
        warnings.warn(
            "DarFile.get_package is deprecated; use DarFile.async_package_service.get_package instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.package_bytes(package_id)

    async def list_package_ids(self):
        """
        This method is deprecated; use :meth:`DarFile.async_package_service` instead.
        """
        warnings.warn(
            "DarFile.list_package_ids is deprecated; use DarFile.async_package_service.list_package_ids instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.package_ids()

    def _dalf_names(self) -> "Generator[PackageRef, None, None]":
        """
        Return a generator over the names of DALF files in this DarFile.
        """
        return (PackageRef(name) for name in self._zip.namelist() if name.endswith(".dalf"))

    def _pb_archives(self) -> "Generator[pb.Archive, None, None]":
        """
        Return a generator over :class:`pb.Archive` instances. Crucially, the Protobuf messages
        contain DAML-LF as a ``bytes`` field that has not yet been parsed.
        """
        for name in self._dalf_names():
            contents = self._zip.read(name)

            a = pb.Archive()
            a.ParseFromString(contents)
            yield a


class CachedDarFile:
    """
    A caching variation of :class:`DarFile`.

    :class:`CachedDarFile` tries to do disk operations only once, and hold onto their results in
    memory. This will generally improve performance at the expense of increased memory usage.
    """

    def __init__(self, dar: "Dar"):
        self._dar = dar
        from threading import Lock

        self._lock = Lock()
        self._archives = None  # type: Optional[Collection[Archive]]

    def archives(self) -> "Collection[Archive]":
        if self._archives is None:
            with self._lock:
                if self._archives is None:
                    with DarFile(self._dar) as dar:
                        self._archives = dar.archives()
        return self._archives

    def package(self, package_id: "PackageRef") -> "Package":
        # TODO: This import needs to be local as long as the dazl.util.dar module still exists
        #  to avoid import cycles. Move this to the top of the file when dazl.util.dar is removed.
        from .errors import PackageNotFoundError

        for archive in self.archives():
            if archive.hash == package_id:
                return archive.package

        raise PackageNotFoundError(package_id)

    def package_ids(self) -> "AbstractSet[PackageRef]":
        return frozenset([archive.hash for archive in self.archives()])


def get_dar_package_ids(dar: "Dar") -> "AbstractSet[PackageRef]":
    """
    Return the package IDs for a DAR.
    """
    with DarFile(dar) as dar_file:
        return dar_file.package_ids()


class DarFileAsyncPackageService:
    def __init__(self, __dar_file: "DarFile"):
        self._dar_file = __dar_file

    async def get_package(self, package_id: "PackageRef") -> bytes:
        return self._dar_file.package_bytes(package_id)

    async def list_package_ids(self) -> "AbstractSet[PackageRef]":
        return self._dar_file.package_ids()


class DarFileBlockingPackageService:
    def __init__(self, __dar_file: "DarFile"):
        self._dar_file = __dar_file
        self._lock = threading.RLock()

    def get_package(self, package_id: "PackageRef") -> bytes:
        with self._lock:
            return self._dar_file.package_bytes(package_id)

    def list_package_ids(self) -> "AbstractSet[PackageRef]":
        with self._lock:
            return self._dar_file.package_ids()



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, typing as _typing

import grpc as _grpc
from grpc import aio as _grpc_aio

from .party_management_service_pb2 import AllocatePartyRequest, AllocatePartyResponse, GetParticipantIdRequest, GetParticipantIdResponse, GetPartiesRequest, GetPartiesResponse, ListKnownPartiesRequest, ListKnownPartiesResponse

__all__ = [
    "PartyManagementServiceStub",
]


# noinspection PyPep8Naming,DuplicatedCode
class PartyManagementServiceStub:
    @classmethod
    @_typing.overload
    def __new__(cls, channel: _grpc.Channel) -> _PartyManagementServiceBlockingStub: ...  # type: ignore
    @classmethod
    @_typing.overload
    def __new__(cls, channel: _grpc_aio.Channel) -> _PartyManagementServiceAsyncStub: ...  # type: ignore
    def GetParticipantId(self, __1: GetParticipantIdRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[GetParticipantIdResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, GetParticipantIdResponse]]: ...
    def GetParties(self, __1: GetPartiesRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[GetPartiesResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, GetPartiesResponse]]: ...
    def ListKnownParties(self, __1: ListKnownPartiesRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[ListKnownPartiesResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, ListKnownPartiesResponse]]: ...
    def AllocateParty(self, __1: AllocatePartyRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[AllocatePartyResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, AllocatePartyResponse]]: ...

# noinspection PyPep8Naming,DuplicatedCode
class _PartyManagementServiceBlockingStub(PartyManagementServiceStub):
    def GetParticipantId(self, __1: GetParticipantIdRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> GetParticipantIdResponse: ...
    def GetParties(self, __1: GetPartiesRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> GetPartiesResponse: ...
    def ListKnownParties(self, __1: ListKnownPartiesRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> ListKnownPartiesResponse: ...
    def AllocateParty(self, __1: AllocatePartyRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> AllocatePartyResponse: ...

# noinspection PyPep8Naming,DuplicatedCode
class _PartyManagementServiceAsyncStub(PartyManagementServiceStub):
    def GetParticipantId(self, __1: GetParticipantIdRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, GetParticipantIdResponse]: ...
    def GetParties(self, __1: GetPartiesRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, GetPartiesResponse]: ...
    def ListKnownParties(self, __1: ListKnownPartiesRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, ListKnownPartiesResponse]: ...
    def AllocateParty(self, __1: AllocatePartyRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, AllocatePartyResponse]: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains supporting infrastructure for built-in method definitions for interpreting
Daml-LF files.
"""

from typing import Any, Optional, Sequence, Type as PyType, Union
import warnings

from .daml_lf_1 import BuiltinFunction, Expr, Type, ValName
from .util import package_local_name

with warnings.catch_warnings():
    warnings.simplefilter("ignore", DeprecationWarning)
    from ..model.types import TypeReference


class _BuiltinMeta(type):
    def __new__(mcs, name, bases, dct):
        if name != "Builtin":
            proposed_name = dct.get("name", None)
            proposed_builtin = dct.get("builtin", None)
            if proposed_name is None:
                if proposed_builtin is None:
                    raise ValueError("Neither `name` nor `builtin` are specified")
            elif proposed_builtin is not None:
                raise ValueError("Only `name` or `builtin` can be specified, not both")
            dct["name"] = proposed_name
            dct["builtin"] = proposed_builtin
            args = dct.get("args")
            if args is None:
                raise ValueError(f"`args` must be specified for class {proposed_name}")
            if isinstance(args, Type):
                dct["args"] = (args,)
            elif not isinstance(args, tuple):
                dct["args"] = tuple(args)
        return super().__new__(mcs, name, bases, dct)


class Builtin(metaclass=_BuiltinMeta):
    name: Optional[str]
    builtin: Optional[BuiltinFunction]
    arg_types: Sequence[Type]

    def evaluate(self, type_args: "Sequence[Type]", args: "Sequence[Any]") -> "Any":
        raise NotImplementedError()

    def simplify(self, type_args: "Sequence[Type]", args: "Sequence[Expr]") -> "Optional[Expr]":
        raise NotImplementedError()


class BuiltinTable:
    def __init__(self):
        self.by_name = dict()  # type: [BuiltinFunction, PyType[Builtin]]
        self.by_builtin = dict()  # type: [str, PyType[Builtin]]

    def add(self, builtin: "PyType[Builtin]"):
        if builtin.builtin is not None:
            self.by_builtin[builtin.builtin] = builtin
        elif builtin.name is not None:
            self.by_name[builtin.name] = builtin
        else:
            raise ValueError(f"A builtin could not be registered! {builtin!r}")

    def resolve(
        self, ref: "Union[str, ValName, TypeReference, BuiltinFunction]"
    ) -> "Optional[Builtin]":
        """
        Return a :class:`Builtin` implementation for the name or reference if one is defined.
        """
        if isinstance(ref, BuiltinFunction):
            # All BuiltinFunctions MUST be defined
            return self.by_builtin[ref]
        elif isinstance(ref, (ValName, TypeReference)):
            return self.by_name.get(package_local_name(ref))
        elif isinstance(ref, str):
            return self.by_name.get(ref)
        else:
            raise TypeError(f"unexpected key type: {ref!r}")

    def __contains__(self, ref):
        if isinstance(ref, BuiltinFunction):
            return True



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from google.protobuf.compiler.plugin_pb2 import CodeGeneratorRequest, CodeGeneratorResponse

from . import plugin_grpc, plugin_grpc_pyi, plugin_init, plugin_pb, plugin_pb_pyi

__all__ = ["run_all_plugins"]


def run_all_plugins(request: "CodeGeneratorRequest") -> "CodeGeneratorResponse":
    response = CodeGeneratorResponse()
    response.MergeFrom(plugin_init.run_plugin(request))
    response.MergeFrom(plugin_grpc.run_plugin(request))
    response.MergeFrom(plugin_grpc_pyi.run_plugin(request))
    response.MergeFrom(plugin_pb.run_plugin(request))
    response.MergeFrom(plugin_pb_pyi.run_plugin(request))
    return response



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/admin/config_management_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n<com/daml/ledger/api/v1/admin/config_management_service.proto\x12\x1c\x63om.daml.ledger.api.v1.admin\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\x15\n\x13GetTimeModelRequest\"\x99\x01\n\x14GetTimeModelResponse\x12\x39\n\x18\x63onfiguration_generation\x18\x01 \x01(\x03R\x17\x63onfigurationGeneration\x12\x46\n\ntime_model\x18\x02 \x01(\x0b\x32\'.com.daml.ledger.api.v1.admin.TimeModelR\ttimeModel\"\x90\x02\n\x13SetTimeModelRequest\x12#\n\rsubmission_id\x18\x01 \x01(\tR\x0csubmissionId\x12J\n\x13maximum_record_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x11maximumRecordTime\x12\x39\n\x18\x63onfiguration_generation\x18\x03 \x01(\x03R\x17\x63onfigurationGeneration\x12M\n\x0enew_time_model\x18\x04 \x01(\x0b\x32\'.com.daml.ledger.api.v1.admin.TimeModelR\x0cnewTimeModel\"Q\n\x14SetTimeModelResponse\x12\x39\n\x18\x63onfiguration_generation\x18\x01 \x01(\x03R\x17\x63onfigurationGeneration\"\xdc\x01\n\tTimeModel\x12Q\n\x17\x61vg_transaction_latency\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationR\x15\x61vgTransactionLatency\x12\x34\n\x08min_skew\x18\x05 \x01(\x0b\x32\x19.google.protobuf.DurationR\x07minSkew\x12\x34\n\x08max_skew\x18\x06 \x01(\x0b\x32\x19.google.protobuf.DurationR\x07maxSkewJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\x03\x10\x04\x32\x87\x02\n\x17\x43onfigManagementService\x12u\n\x0cGetTimeModel\x12\x31.com.daml.ledger.api.v1.admin.GetTimeModelRequest\x1a\x32.com.daml.ledger.api.v1.admin.GetTimeModelResponse\x12u\n\x0cSetTimeModel\x12\x31.com.daml.ledger.api.v1.admin.SetTimeModelRequest\x1a\x32.com.daml.ledger.api.v1.admin.SetTimeModelResponseB\xad\x01\n\x1c\x63om.daml.ledger.api.v1.adminB!ConfigManagementServiceOuterClassZKgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1/admin\xaa\x02\x1c\x43om.Daml.Ledger.Api.V1.Adminb\x06proto3')



_GETTIMEMODELREQUEST = DESCRIPTOR.message_types_by_name['GetTimeModelRequest']
_GETTIMEMODELRESPONSE = DESCRIPTOR.message_types_by_name['GetTimeModelResponse']
_SETTIMEMODELREQUEST = DESCRIPTOR.message_types_by_name['SetTimeModelRequest']
_SETTIMEMODELRESPONSE = DESCRIPTOR.message_types_by_name['SetTimeModelResponse']
_TIMEMODEL = DESCRIPTOR.message_types_by_name['TimeModel']
GetTimeModelRequest = _reflection.GeneratedProtocolMessageType('GetTimeModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTIMEMODELREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.admin.config_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.GetTimeModelRequest)
  })
_sym_db.RegisterMessage(GetTimeModelRequest)

GetTimeModelResponse = _reflection.GeneratedProtocolMessageType('GetTimeModelResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETTIMEMODELRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.admin.config_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.GetTimeModelResponse)
  })
_sym_db.RegisterMessage(GetTimeModelResponse)

SetTimeModelRequest = _reflection.GeneratedProtocolMessageType('SetTimeModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETTIMEMODELREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.admin.config_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.SetTimeModelRequest)
  })
_sym_db.RegisterMessage(SetTimeModelRequest)

SetTimeModelResponse = _reflection.GeneratedProtocolMessageType('SetTimeModelResponse', (_message.Message,), {
  'DESCRIPTOR' : _SETTIMEMODELRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.admin.config_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.SetTimeModelResponse)
  })
_sym_db.RegisterMessage(SetTimeModelResponse)

TimeModel = _reflection.GeneratedProtocolMessageType('TimeModel', (_message.Message,), {
  'DESCRIPTOR' : _TIMEMODEL,
  '__module__' : 'com.daml.ledger.api.v1.admin.config_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.TimeModel)
  })
_sym_db.RegisterMessage(TimeModel)

_CONFIGMANAGEMENTSERVICE = DESCRIPTOR.services_by_name['ConfigManagementService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\034com.daml.ledger.api.v1.adminB!ConfigManagementServiceOuterClassZKgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1/admin\252\002\034Com.Daml.Ledger.Api.V1.Admin'
  _GETTIMEMODELREQUEST._serialized_start=159
  _GETTIMEMODELREQUEST._serialized_end=180
  _GETTIMEMODELRESPONSE._serialized_start=183
  _GETTIMEMODELRESPONSE._serialized_end=336
  _SETTIMEMODELREQUEST._serialized_start=339
  _SETTIMEMODELREQUEST._serialized_end=611
  _SETTIMEMODELRESPONSE._serialized_start=613
  _SETTIMEMODELRESPONSE._serialized_end=694
  _TIMEMODEL._serialized_start=697
  _TIMEMODEL._serialized_end=917
  _CONFIGMANAGEMENTSERVICE._serialized_start=920
  _CONFIGMANAGEMENTSERVICE._serialized_end=1183
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Methods for serializing domain objects and other things into basic primitive
types over the wire on the REST interface using JSON.
"""

from datetime import datetime
from typing import TYPE_CHECKING, Any
import warnings

from ...damlast.daml_lf_1 import TypeConName
from ...model.writing import AbstractSerializer
from ...prim import ContractId, JSONEncoder
from ...values.json import JsonEncoder

if TYPE_CHECKING:
    from ...model.writing import CommandPayload

__all__ = ["LedgerJSONEncoder", "to_api_datetime", "JsonSerializer"]


class LedgerJSONEncoder(JSONEncoder):
    """
    Convert some known Ledger API primitive types into their appropriate JSON
    representations.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn(
            "dazl.protocols.v0.json_ser_command.LedgerJSONEncoder is deprecated; "
            "use dazl.prim.JSONEncoder instead.",
            DeprecationWarning,
            stacklevel=2,
        )


def to_api_datetime(obj):
    """
    Converts the object to an ISO8601 datetime string.

    :param obj:
        A datetime. If the datetime is "naive" (no timezone information), it is
        assumed to refer to UTC. If the datetime has timezone information, the
        datetime will be converted to UTC before serializing.
    :return: An ISO8601 string that represents the time in UTC.
    """
    from ...prim import datetime_to_str

    warnings.warn(
        "dazl.protocols.v0.json_ser_command.to_api_datetime is deprecated; "
        "use dazl.prim.datetime_to_str instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return datetime_to_str(obj) if isinstance(obj, datetime) else obj


class JsonSerializer(AbstractSerializer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn(
            "dazl.protocols.v0.json_ser_command.JsonSerializer is deprecated; "
            "there is no replacement.",
            DeprecationWarning,
            stacklevel=2,
        )

    mapper = JsonEncoder()

    def serialize_command_request(self, command_payload: "CommandPayload") -> dict:
        commands = [self.serialize_command(command) for command in command_payload.commands]
        return dict(
            businessIntent=command_payload.command_id,
            commands=commands,
            application=command_payload.application_id,
        )

    def serialize_create_command(self, template_name: "TypeConName", template_args: "Any") -> "Any":
        # the package_id in ModuleRef is a convenient place to
        # stash legacy template IDs in the REST endpoint
        return {"create": {"template": str(template_name), "arguments": template_args}}

    def serialize_exercise_command(
        self, contract_id: ContractId, choice_name: str, choice_args: "Any"
    ) -> "Any":
        return {
            "exercise": {
                "contract": contract_id.value,
                "choice": choice_name,
                "arguments": choice_args,
            }
        }



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Collection, Optional


class Certificate:
    def __init__(self, public_cert: bytes, private_key: bytes):
        self.public_cert = public_cert
        self.private_key = private_key


def cert_gen(
    email_address: str = "emailAddress",
    common_name: str = "commonName",
    country_name: str = "NT",
    locality_name: str = "localityName",
    state_or_province_name: str = "stateOrProvinceName",
    organization_name: str = "organizationName",
    organization_unit_name: str = "organizationUnitName",
    subject_alternative_name: "Optional[Collection[str]]" = None,
    serial_number: int = 0,
    validity_end_in_seconds: int = 10 * 365 * 24 * 60 * 60,
) -> "Certificate":
    from OpenSSL import crypto

    # can look at generated file using openssl:
    # openssl x509 -inform pem -in selfsigned.crt -noout -text
    # create a key pair
    k = crypto.PKey()
    k.generate_key(crypto.TYPE_RSA, 4096)
    # create a self-signed cert
    cert = crypto.X509()
    cert.set_version(2)
    subject = cert.get_subject()
    subject.C = country_name
    subject.ST = state_or_province_name
    subject.L = locality_name
    subject.O = organization_name
    subject.OU = organization_unit_name
    subject.CN = common_name
    subject.emailAddress = email_address
    cert.set_serial_number(serial_number)
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(validity_end_in_seconds)
    cert.set_issuer(cert.get_subject())
    cert.set_pubkey(k)
    if subject_alternative_name:
        if isinstance(subject_alternative_name, str):
            subject_alternative_name = (subject_alternative_name,)
        cert.add_extensions(
            [
                crypto.X509Extension(
                    b"subjectAltName",
                    False,
                    ",".join(f"DNS:{san}" for san in subject_alternative_name).encode("ascii"),
                )
            ]
        )

    cert.sign(k, "sha512")
    return Certificate(
        public_cert=crypto.dump_certificate(crypto.FILETYPE_PEM, cert),
        private_key=crypto.dump_privatekey(crypto.FILETYPE_PEM, k),
    )



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/ledger_offset.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*com/daml/ledger/api/v1/ledger_offset.proto\x12\x16\x63om.daml.ledger.api.v1\"\xbc\x01\n\x0cLedgerOffset\x12\x1c\n\x08\x61\x62solute\x18\x01 \x01(\tH\x00R\x08\x61\x62solute\x12Q\n\x08\x62oundary\x18\x02 \x01(\x0e\x32\x33.com.daml.ledger.api.v1.LedgerOffset.LedgerBoundaryH\x00R\x08\x62oundary\"2\n\x0eLedgerBoundary\x12\x10\n\x0cLEDGER_BEGIN\x10\x00\x12\x0e\n\nLEDGER_END\x10\x01\x42\x07\n\x05valueB\x90\x01\n\x16\x63om.daml.ledger.api.v1B\x16LedgerOffsetOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3')



_LEDGEROFFSET = DESCRIPTOR.message_types_by_name['LedgerOffset']
_LEDGEROFFSET_LEDGERBOUNDARY = _LEDGEROFFSET.enum_types_by_name['LedgerBoundary']
LedgerOffset = _reflection.GeneratedProtocolMessageType('LedgerOffset', (_message.Message,), {
  'DESCRIPTOR' : _LEDGEROFFSET,
  '__module__' : 'com.daml.ledger.api.v1.ledger_offset_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.LedgerOffset)
  })
_sym_db.RegisterMessage(LedgerOffset)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.daml.ledger.api.v1B\026LedgerOffsetOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\252\002\026Com.Daml.Ledger.Api.V1'
  _LEDGEROFFSET._serialized_start=71
  _LEDGEROFFSET._serialized_end=259
  _LEDGEROFFSET_LEDGERBOUNDARY._serialized_start=200
  _LEDGEROFFSET_LEDGERBOUNDARY._serialized_end=250
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Formatting module for outputting captures.
"""

import json
from typing import AbstractSet, Iterable

from ...damlast.daml_types import con
from ...damlast.protocols import SymbolLookup
from ...prim import Party
from ...util.tools import boundary_iter
from ...values import Context, JsonEncoder
from .model import Formatter, RowBuilder


class JsonFormatter(Formatter):
    def render(
        self, lookup: "SymbolLookup", parties: "AbstractSet[Party]", entries: "Iterable[RowBuilder]"
    ):
        """
        Return a list of strings that, when taken together, constitute a JSON document that lists
        all of the entries of the ledger.

        :param lookup:
            The contents to render.
        :param parties:
            The list of of parties.
        :param entries:
            The entries to render.
        """
        context = Context(JsonEncoder(), lookup)

        party_list = sorted(parties)

        yield '{{ "parties": {},'.format(json.dumps(party_list))
        yield '  "contracts": ['

        for is_last, entry in boundary_iter(entries):
            cid = entry.cid
            yield f'  {{ "contract_id": "{cid.value}",'
            yield f'    "template": {json.dumps(str(cid.value_type))},'
            yield f'    "parties": {json.dumps(party_list)},'
            yield f'    "arguments": {json.dumps(context.convert(con(cid.value_type), entry.cdata))} }}' + (
                "" if is_last else ","
            )

        yield "  ] }"



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, typing as _typing

import grpc as _grpc
from grpc import aio as _grpc_aio

from .transaction_service_pb2 import GetFlatTransactionResponse, GetLedgerEndRequest, GetLedgerEndResponse, GetTransactionByEventIdRequest, GetTransactionByIdRequest, GetTransactionResponse, GetTransactionTreesResponse, GetTransactionsRequest, GetTransactionsResponse

__all__ = [
    "TransactionServiceStub",
]


# noinspection PyPep8Naming,DuplicatedCode
class TransactionServiceStub:
    @classmethod
    @_typing.overload
    def __new__(cls, channel: _grpc.Channel) -> _TransactionServiceBlockingStub: ...  # type: ignore
    @classmethod
    @_typing.overload
    def __new__(cls, channel: _grpc_aio.Channel) -> _TransactionServiceAsyncStub: ...  # type: ignore
    def GetTransactions(self, __1: GetTransactionsRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[_grpc.CallIterator[GetTransactionsResponse], _grpc_aio.UnaryStreamCall[_typing.Any, GetTransactionsResponse]]: ...
    def GetTransactionTrees(self, __1: GetTransactionsRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[_grpc.CallIterator[GetTransactionTreesResponse], _grpc_aio.UnaryStreamCall[_typing.Any, GetTransactionTreesResponse]]: ...
    def GetTransactionByEventId(self, __1: GetTransactionByEventIdRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[GetTransactionResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, GetTransactionResponse]]: ...
    def GetTransactionById(self, __1: GetTransactionByIdRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[GetTransactionResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, GetTransactionResponse]]: ...
    def GetFlatTransactionByEventId(self, __1: GetTransactionByEventIdRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[GetFlatTransactionResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, GetFlatTransactionResponse]]: ...
    def GetFlatTransactionById(self, __1: GetTransactionByIdRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[GetFlatTransactionResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, GetFlatTransactionResponse]]: ...
    def GetLedgerEnd(self, __1: GetLedgerEndRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _typing.Union[GetLedgerEndResponse, _grpc_aio.UnaryUnaryCall[_typing.Any, GetLedgerEndResponse]]: ...

# noinspection PyPep8Naming,DuplicatedCode
class _TransactionServiceBlockingStub(TransactionServiceStub):
    def GetTransactions(self, __1: GetTransactionsRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc.CallIterator[GetTransactionsResponse]: ...
    def GetTransactionTrees(self, __1: GetTransactionsRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc.CallIterator[GetTransactionTreesResponse]: ...
    def GetTransactionByEventId(self, __1: GetTransactionByEventIdRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> GetTransactionResponse: ...
    def GetTransactionById(self, __1: GetTransactionByIdRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> GetTransactionResponse: ...
    def GetFlatTransactionByEventId(self, __1: GetTransactionByEventIdRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> GetFlatTransactionResponse: ...
    def GetFlatTransactionById(self, __1: GetTransactionByIdRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> GetFlatTransactionResponse: ...
    def GetLedgerEnd(self, __1: GetLedgerEndRequest, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> GetLedgerEndResponse: ...

# noinspection PyPep8Naming,DuplicatedCode
class _TransactionServiceAsyncStub(TransactionServiceStub):
    def GetTransactions(self, __1: GetTransactionsRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryStreamCall[_typing.Any, GetTransactionsResponse]: ...
    def GetTransactionTrees(self, __1: GetTransactionsRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryStreamCall[_typing.Any, GetTransactionTreesResponse]: ...
    def GetTransactionByEventId(self, __1: GetTransactionByEventIdRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, GetTransactionResponse]: ...
    def GetTransactionById(self, __1: GetTransactionByIdRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, GetTransactionResponse]: ...
    def GetFlatTransactionByEventId(self, __1: GetTransactionByEventIdRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, GetFlatTransactionResponse]: ...
    def GetFlatTransactionById(self, __1: GetTransactionByIdRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, GetFlatTransactionResponse]: ...
    def GetLedgerEnd(self, __1: GetLedgerEndRequest, *, timeout: _typing.Optional[float] = ..., metadata: _typing.Optional[_grpc_aio.Metadata] = ..., credentials: _typing.Optional[_grpc.CallCredentials] = ..., wait_for_ready: _typing.Optional[bool] = ..., compression: _typing.Optional[_grpc.Compression] = ...) -> _grpc_aio.UnaryUnaryCall[_typing.Any, GetLedgerEndResponse]: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Compatibility methods
---------------------

These functions should only be used internally by code that is helping in the transition from the
deprecated :class:`dazl.model.types.Type` hierarchy to :class:`dazl.damlast.daml_lf_1.Type`.
These functions will be removed WITHOUT an intermediate deprecation warning, as they are considered
internal only.
"""
from typing import TYPE_CHECKING, Tuple, Union
import warnings

from .daml_lf_1 import TypeConName

if TYPE_CHECKING:
    # avoid import cycles; `dazl.model` should depend on `dazl.damlast`, and not the other way
    # around
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        from ..model.types import Type as DeprecatedType, TypeReference as DeprecatedTypeReference


def parse_template(
    template_id: "Union[str, DeprecatedType, TypeConName]",
) -> "Tuple[TypeConName, DeprecatedTypeReference]":
    """
    Return both the "new-style" Type and the "old-style" Type for _either_ a string, "new-style"
    Type, or "old-style" Type. This is used in places where we can't easily mark a symbol as
    deprecated (particularly, constructors for types that are NOT to be deprecated as part of this
    transition).
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        from ..model.types import (
            Type as DeprecatedType,
            TypeReference as DeprecatedTypeReference,
            UnresolvedTypeReference,
            RecordType,
        )
    from ..damlast.lookup import parse_type_con_name

    if isinstance(template_id, str):
        name = parse_type_con_name(template_id, allow_deprecated_identifiers=True)
        return name, DeprecatedTypeReference(name)

    elif isinstance(template_id, TypeConName):
        return template_id, DeprecatedTypeReference(template_id)

    elif isinstance(template_id, DeprecatedType):
        warnings.warn(
            "usage of dazl.model.types.Type is deprecated; use TypeConName instead",
            DeprecationWarning,
            stacklevel=3,
        )

        if isinstance(template_id, DeprecatedTypeReference):
            return template_id.con, template_id

        elif isinstance(template_id, UnresolvedTypeReference):
            name = parse_type_con_name(template_id.name, allow_deprecated_identifiers=True)
            return name, DeprecatedTypeReference(name)

        elif isinstance(template_id, RecordType):
            if template_id.name is None:
                raise ValueError(
                    "template_id must point to a named record that corresponds to a template"
                )

            return template_id.name.con, DeprecatedTypeReference(template_id.name.con)
        else:
            raise ValueError("unknown dazl.model.types.Type implementation")
    else:
        raise ValueError("template_id must be a TypeConName")



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import TypeVar


class _Missing:
    """
    Sentinel object used to denote that a value was unspecified in a sum type in the AST.
    """

    __slots__ = ()

    def __repr__(self):
        return "MISSING"


MISSING = _Missing()
T = TypeVar("T")



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.duration_pb2 import Duration
from google.protobuf.internal.containers import RepeatedScalarFieldContainer
from google.protobuf.message import Message as _Message
from google.rpc.status_pb2 import Status

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "Completion",
]


class Completion(_Message):
    command_id: _builtins.str
    @property
    def status(self) -> Status: ...
    transaction_id: _builtins.str
    application_id: _builtins.str
    @property
    def act_as(self) -> RepeatedScalarFieldContainer[_builtins.str]: ...
    submission_id: _builtins.str
    deduplication_offset: _builtins.str
    @property
    def deduplication_duration(self) -> Duration: ...
    @_typing.overload
    def __init__(self, *, command_id: _typing.Optional[_builtins.str] = ..., status: _typing.Optional[Status] = ..., transaction_id: _typing.Optional[_builtins.str] = ..., application_id: _typing.Optional[_builtins.str] = ..., act_as: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., submission_id: _typing.Optional[_builtins.str] = ...): ...
    @_typing.overload
    def __init__(self, *, command_id: _typing.Optional[_builtins.str] = ..., status: _typing.Optional[Status] = ..., transaction_id: _typing.Optional[_builtins.str] = ..., application_id: _typing.Optional[_builtins.str] = ..., act_as: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., submission_id: _typing.Optional[_builtins.str] = ..., deduplication_offset: _builtins.str = ...): ...
    @_typing.overload
    def __init__(self, *, command_id: _typing.Optional[_builtins.str] = ..., status: _typing.Optional[Status] = ..., transaction_id: _typing.Optional[_builtins.str] = ..., application_id: _typing.Optional[_builtins.str] = ..., act_as: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., submission_id: _typing.Optional[_builtins.str] = ..., deduplication_duration: Duration = ...): ...
    def HasField(self, field_name: _L["command_id", "status", "transaction_id", "application_id", "act_as", "submission_id", "deduplication_period", "deduplication_offset", "deduplication_duration"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["command_id", "status", "transaction_id", "application_id", "act_as", "submission_id", "deduplication_period", "deduplication_offset", "deduplication_duration"]) -> None: ...
    def WhichOneof(self, oneof_group: _L["deduplication_period"]) -> _L[None, "deduplication_offset", "deduplication_duration"]: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module has been relocated to ``dazl.client.commands``, though if possible you should move to
the command types defined in ``dazl.ledger`` instead.
"""
import warnings

from ..client.commands import (
    CommandBuilder,
    CommandDefaults,
    CommandPayload,
    CommandsOrCommandSequence,
    CreateAndExerciseCommand,
    CreateCommand,
    EventHandlerResponse,
    ExerciseByKeyCommand,
    ExerciseCommand,
    create,
    create_and_exercise,
    exercise,
    exercise_by_key,
    flatten_command_sequence,
)
from ..ledger import Command
from ..protocols.serializers import AbstractSerializer, Serializer

warnings.warn(
    "dazl.model.writing is deprecated; please use the types of dazl.ledger.api_types instead.",
    DeprecationWarning,
    stacklevel=2,
)

__all__ = [
    "AbstractSerializer",
    "Command",
    "CommandBuilder",
    "CommandDefaults",
    "CommandPayload",
    "CommandsOrCommandSequence",
    "CreateAndExerciseCommand",
    "CreateCommand",
    "EventHandlerResponse",
    "ExerciseByKeyCommand",
    "ExerciseCommand",
    "Serializer",
    "create",
    "create_and_exercise",
    "exercise",
    "exercise_by_key",
    "flatten_command_sequence",
]



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/command_completion_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import completion_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_completion__pb2
from . import ledger_offset_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__offset__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n7com/daml/ledger/api/v1/command_completion_service.proto\x12\x16\x63om.daml.ledger.api.v1\x1a\'com/daml/ledger/api/v1/completion.proto\x1a*com/daml/ledger/api/v1/ledger_offset.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xb5\x01\n\x17\x43ompletionStreamRequest\x12\x1b\n\tledger_id\x18\x01 \x01(\tR\x08ledgerId\x12%\n\x0e\x61pplication_id\x18\x02 \x01(\tR\rapplicationId\x12\x18\n\x07parties\x18\x03 \x03(\tR\x07parties\x12<\n\x06offset\x18\x04 \x01(\x0b\x32$.com.daml.ledger.api.v1.LedgerOffsetR\x06offset\"\xa4\x01\n\x18\x43ompletionStreamResponse\x12\x42\n\ncheckpoint\x18\x01 \x01(\x0b\x32\".com.daml.ledger.api.v1.CheckpointR\ncheckpoint\x12\x44\n\x0b\x63ompletions\x18\x02 \x03(\x0b\x32\".com.daml.ledger.api.v1.CompletionR\x0b\x63ompletions\"\x87\x01\n\nCheckpoint\x12;\n\x0brecord_time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\nrecordTime\x12<\n\x06offset\x18\x02 \x01(\x0b\x32$.com.daml.ledger.api.v1.LedgerOffsetR\x06offset\"3\n\x14\x43ompletionEndRequest\x12\x1b\n\tledger_id\x18\x01 \x01(\tR\x08ledgerId\"U\n\x15\x43ompletionEndResponse\x12<\n\x06offset\x18\x01 \x01(\x0b\x32$.com.daml.ledger.api.v1.LedgerOffsetR\x06offset2\x81\x02\n\x18\x43ommandCompletionService\x12w\n\x10\x43ompletionStream\x12/.com.daml.ledger.api.v1.CompletionStreamRequest\x1a\x30.com.daml.ledger.api.v1.CompletionStreamResponse0\x01\x12l\n\rCompletionEnd\x12,.com.daml.ledger.api.v1.CompletionEndRequest\x1a-.com.daml.ledger.api.v1.CompletionEndResponseB\x9c\x01\n\x16\x63om.daml.ledger.api.v1B\"CommandCompletionServiceOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3')



_COMPLETIONSTREAMREQUEST = DESCRIPTOR.message_types_by_name['CompletionStreamRequest']
_COMPLETIONSTREAMRESPONSE = DESCRIPTOR.message_types_by_name['CompletionStreamResponse']
_CHECKPOINT = DESCRIPTOR.message_types_by_name['Checkpoint']
_COMPLETIONENDREQUEST = DESCRIPTOR.message_types_by_name['CompletionEndRequest']
_COMPLETIONENDRESPONSE = DESCRIPTOR.message_types_by_name['CompletionEndResponse']
CompletionStreamRequest = _reflection.GeneratedProtocolMessageType('CompletionStreamRequest', (_message.Message,), {
  'DESCRIPTOR' : _COMPLETIONSTREAMREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.command_completion_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.CompletionStreamRequest)
  })
_sym_db.RegisterMessage(CompletionStreamRequest)

CompletionStreamResponse = _reflection.GeneratedProtocolMessageType('CompletionStreamResponse', (_message.Message,), {
  'DESCRIPTOR' : _COMPLETIONSTREAMRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.command_completion_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.CompletionStreamResponse)
  })
_sym_db.RegisterMessage(CompletionStreamResponse)

Checkpoint = _reflection.GeneratedProtocolMessageType('Checkpoint', (_message.Message,), {
  'DESCRIPTOR' : _CHECKPOINT,
  '__module__' : 'com.daml.ledger.api.v1.command_completion_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Checkpoint)
  })
_sym_db.RegisterMessage(Checkpoint)

CompletionEndRequest = _reflection.GeneratedProtocolMessageType('CompletionEndRequest', (_message.Message,), {
  'DESCRIPTOR' : _COMPLETIONENDREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.command_completion_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.CompletionEndRequest)
  })
_sym_db.RegisterMessage(CompletionEndRequest)

CompletionEndResponse = _reflection.GeneratedProtocolMessageType('CompletionEndResponse', (_message.Message,), {
  'DESCRIPTOR' : _COMPLETIONENDRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.command_completion_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.CompletionEndResponse)
  })
_sym_db.RegisterMessage(CompletionEndResponse)

_COMMANDCOMPLETIONSERVICE = DESCRIPTOR.services_by_name['CommandCompletionService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.daml.ledger.api.v1B\"CommandCompletionServiceOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\252\002\026Com.Daml.Ledger.Api.V1'
  _COMPLETIONSTREAMREQUEST._serialized_start=202
  _COMPLETIONSTREAMREQUEST._serialized_end=383
  _COMPLETIONSTREAMRESPONSE._serialized_start=386
  _COMPLETIONSTREAMRESPONSE._serialized_end=550
  _CHECKPOINT._serialized_start=553
  _CHECKPOINT._serialized_end=688
  _COMPLETIONENDREQUEST._serialized_start=690
  _COMPLETIONENDREQUEST._serialized_end=741
  _COMPLETIONENDRESPONSE._serialized_start=743
  _COMPLETIONENDRESPONSE._serialized_end=828
  _COMMANDCOMPLETIONSERVICE._serialized_start=831
  _COMMANDCOMPLETIONSERVICE._serialized_end=1088
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from asyncio import new_event_loop, run_coroutine_threadsafe, set_event_loop, sleep
from queue import Queue
from threading import Thread
from typing import Callable, Optional

from ..aio import Connection, QueryStream

__all__ = ["ConnectionThunk", "QueryStreamThunk"]


class ConnectionThunk:
    """
    A blocking Connection wrapper around an asynchronous connection.
    """

    def __init__(self, conn_fn: "Callable[[], Connection]", *, name: "Optional[str]" = None):
        self._conn_fn = conn_fn
        self._conn = None
        self._thread = Thread(target=self._main, name=name, daemon=True)
        self._loop = new_event_loop()
        self._fut = self._loop.create_future()

    def _main(self) -> None:
        """
        The "main" thread that runs the event loop where operations for this connection are scheduled.
        """
        set_event_loop(self._loop)
        self._loop.run_until_complete(self._fut)

    def open(self):
        # start the main thread
        self._thread.start()

        fut = run_coroutine_threadsafe(self._create_conn(), self._loop)
        fut.result()

    async def _create_conn(self):
        await sleep(0)
        self._conn = self._conn_fn()
        await self._conn.open()

    def close(self):
        run_coroutine_threadsafe(self._conn.close(), self._loop).result()
        self._loop.call_soon_threadsafe(self._fut.set_result, None)
        self._thread.join()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def create(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.create(*args, **kwargs), self._loop)
        return fut.result()

    def create_and_exercise(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.create_and_exercise(*args, **kwargs), self._loop)
        return fut.result()

    def exercise(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.exercise(*args, **kwargs), self._loop)
        return fut.result()

    def exercise_by_key(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.exercise_by_key(*args, **kwargs), self._loop)
        return fut.result()

    def submit(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.submit(*args, **kwargs), self._loop)
        return fut.result()

    def archive(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.archive(*args, **kwargs), self._loop)
        return fut.result()

    def archive_by_key(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.archive_by_key(*args, **kwargs), self._loop)
        return fut.result()

    def get_ledger_end(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.get_ledger_end(*args, **kwargs), self._loop)
        return fut.result()

    def query(self, *args, **kwargs):
        return QueryStreamThunk(self._conn.query(*args, **kwargs), self._loop)

    def query_many(self, *args, **kwargs):
        return QueryStreamThunk(self._conn.query_many(*args, **kwargs), self._loop)

    def stream(self, *args, **kwargs):
        return QueryStreamThunk(self._conn.stream(*args, **kwargs), self._loop)

    def stream_many(self, *args, **kwargs):
        return QueryStreamThunk(self._conn.stream_many(*args, **kwargs), self._loop)

    def get_user(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.get_user(*args, **kwargs), self._loop)
        return fut.result()

    def list_users(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.list_users(*args, **kwargs), self._loop)
        return fut.result()

    def allocate_party(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.allocate_party(*args, **kwargs), self._loop)
        return fut.result()

    def list_known_parties(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.list_known_parties(*args, **kwargs), self._loop)
        return fut.result()

    def upload_package(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.upload_package(*args, **kwargs), self._loop)
        return fut.result()

    def get_package(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.get_package(*args, **kwargs), self._loop)
        return fut.result()

    def list_package_ids(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.list_package_ids(*args, **kwargs), self._loop)
        return fut.result()

    def get_metering_report(self, *args, **kwargs):
        fut = run_coroutine_threadsafe(self._conn.get_metering_report(*args, **kwargs), self._loop)
        return fut.result()


class QueryStreamThunk:
    """
    A blocking QueryStream wrapper around an asynchronous query stream.
    """

    def __init__(self, stream: "QueryStream", loop):
        self._loop = loop
        self._stream = stream
        self._q = Queue()  # type: ignore
        self._fut = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # try to abort an existing iteration if one is running
        if self._fut is not None:
            try:
                self._fut.cancel()
            except Exception:  # noqa
                pass
        self.close()

    def on_create(self, *args):
        return self._stream.on_create(*args)

    def on_archive(self, *args):
        return self._stream.on_archive(*args)

    def on_boundary(self, *args):
        return self._stream.on_boundary(*args)

    def close(self):
        self._q.put_nowait(None)

    def run(self):
        self._fut = run_coroutine_threadsafe(self._stream.run(), self._loop)
        self._fut.result()

    def creates(self):
        self._fut = run_coroutine_threadsafe(self._producer(self._stream.creates), self._loop)
        return self._consume_queue()

    def events(self):
        self._fut = run_coroutine_threadsafe(self._producer(self._stream.events), self._loop)
        return self._consume_queue()

    def items(self):
        self._fut = run_coroutine_threadsafe(self._producer(self._stream.items), self._loop)
        return self._consume_queue()

    def __iter__(self):
        return self.items()

    async def _producer(self, source):
        async with self._stream:
            async for item in source():
                self._q.put_nowait(item)
            self._q.put_nowait(None)

    def _consume_queue(self):
        while True:
            item = self._q.get()
            if item is not None:
                yield item
            else:
                break



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import ledger_identity_service_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__identity__service__pb2


class LedgerIdentityServiceStub(object):
    """DEPRECATED: This service is now deprecated and ledger identity string is optional for all Ledger API requests.

    Allows clients to verify that the server they are communicating with exposes the ledger they wish to operate on.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetLedgerIdentity = channel.unary_unary(
                '/com.daml.ledger.api.v1.LedgerIdentityService/GetLedgerIdentity',
                request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__identity__service__pb2.GetLedgerIdentityRequest.SerializeToString,
                response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__identity__service__pb2.GetLedgerIdentityResponse.FromString,
                )


class LedgerIdentityServiceServicer(object):
    """DEPRECATED: This service is now deprecated and ledger identity string is optional for all Ledger API requests.

    Allows clients to verify that the server they are communicating with exposes the ledger they wish to operate on.
    """

    def GetLedgerIdentity(self, request, context):
        """Clients may call this RPC to return the identifier of the ledger they are connected to.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_LedgerIdentityServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetLedgerIdentity': grpc.unary_unary_rpc_method_handler(
                    servicer.GetLedgerIdentity,
                    request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__identity__service__pb2.GetLedgerIdentityRequest.FromString,
                    response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__identity__service__pb2.GetLedgerIdentityResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'com.daml.ledger.api.v1.LedgerIdentityService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class LedgerIdentityService(object):
    """DEPRECATED: This service is now deprecated and ledger identity string is optional for all Ledger API requests.

    Allows clients to verify that the server they are communicating with exposes the ledger they wish to operate on.
    """

    @staticmethod
    def GetLedgerIdentity(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/com.daml.ledger.api.v1.LedgerIdentityService/GetLedgerIdentity',
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__identity__service__pb2.GetLedgerIdentityRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_ledger__identity__service__pb2.GetLedgerIdentityResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/admin/package_management_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n=com/daml/ledger/api/v1/admin/package_management_service.proto\x12\x1c\x63om.daml.ledger.api.v1.admin\x1a\x1fgoogle/protobuf/timestamp.proto\"\x1a\n\x18ListKnownPackagesRequest\"r\n\x19ListKnownPackagesResponse\x12U\n\x0fpackage_details\x18\x01 \x03(\x0b\x32,.com.daml.ledger.api.v1.admin.PackageDetailsR\x0epackageDetails\"\xbe\x01\n\x0ePackageDetails\x12\x1d\n\npackage_id\x18\x01 \x01(\tR\tpackageId\x12!\n\x0cpackage_size\x18\x02 \x01(\x04R\x0bpackageSize\x12;\n\x0bknown_since\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\nknownSince\x12-\n\x12source_description\x18\x04 \x01(\tR\x11sourceDescription\"V\n\x14UploadDarFileRequest\x12\x19\n\x08\x64\x61r_file\x18\x01 \x01(\x0cR\x07\x64\x61rFile\x12#\n\rsubmission_id\x18\x02 \x01(\tR\x0csubmissionId\"\x17\n\x15UploadDarFileResponse2\x9b\x02\n\x18PackageManagementService\x12\x84\x01\n\x11ListKnownPackages\x12\x36.com.daml.ledger.api.v1.admin.ListKnownPackagesRequest\x1a\x37.com.daml.ledger.api.v1.admin.ListKnownPackagesResponse\x12x\n\rUploadDarFile\x12\x32.com.daml.ledger.api.v1.admin.UploadDarFileRequest\x1a\x33.com.daml.ledger.api.v1.admin.UploadDarFileResponseB\xae\x01\n\x1c\x63om.daml.ledger.api.v1.adminB\"PackageManagementServiceOuterClassZKgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1/admin\xaa\x02\x1c\x43om.Daml.Ledger.Api.V1.Adminb\x06proto3')



_LISTKNOWNPACKAGESREQUEST = DESCRIPTOR.message_types_by_name['ListKnownPackagesRequest']
_LISTKNOWNPACKAGESRESPONSE = DESCRIPTOR.message_types_by_name['ListKnownPackagesResponse']
_PACKAGEDETAILS = DESCRIPTOR.message_types_by_name['PackageDetails']
_UPLOADDARFILEREQUEST = DESCRIPTOR.message_types_by_name['UploadDarFileRequest']
_UPLOADDARFILERESPONSE = DESCRIPTOR.message_types_by_name['UploadDarFileResponse']
ListKnownPackagesRequest = _reflection.GeneratedProtocolMessageType('ListKnownPackagesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTKNOWNPACKAGESREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.admin.package_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.ListKnownPackagesRequest)
  })
_sym_db.RegisterMessage(ListKnownPackagesRequest)

ListKnownPackagesResponse = _reflection.GeneratedProtocolMessageType('ListKnownPackagesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTKNOWNPACKAGESRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.admin.package_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.ListKnownPackagesResponse)
  })
_sym_db.RegisterMessage(ListKnownPackagesResponse)

PackageDetails = _reflection.GeneratedProtocolMessageType('PackageDetails', (_message.Message,), {
  'DESCRIPTOR' : _PACKAGEDETAILS,
  '__module__' : 'com.daml.ledger.api.v1.admin.package_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.PackageDetails)
  })
_sym_db.RegisterMessage(PackageDetails)

UploadDarFileRequest = _reflection.GeneratedProtocolMessageType('UploadDarFileRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADDARFILEREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.admin.package_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.UploadDarFileRequest)
  })
_sym_db.RegisterMessage(UploadDarFileRequest)

UploadDarFileResponse = _reflection.GeneratedProtocolMessageType('UploadDarFileResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADDARFILERESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.admin.package_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.UploadDarFileResponse)
  })
_sym_db.RegisterMessage(UploadDarFileResponse)

_PACKAGEMANAGEMENTSERVICE = DESCRIPTOR.services_by_name['PackageManagementService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\034com.daml.ledger.api.v1.adminB\"PackageManagementServiceOuterClassZKgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1/admin\252\002\034Com.Daml.Ledger.Api.V1.Admin'
  _LISTKNOWNPACKAGESREQUEST._serialized_start=128
  _LISTKNOWNPACKAGESREQUEST._serialized_end=154
  _LISTKNOWNPACKAGESRESPONSE._serialized_start=156
  _LISTKNOWNPACKAGESRESPONSE._serialized_end=270
  _PACKAGEDETAILS._serialized_start=273
  _PACKAGEDETAILS._serialized_end=463
  _UPLOADDARFILEREQUEST._serialized_start=465
  _UPLOADDARFILEREQUEST._serialized_end=551
  _UPLOADDARFILERESPONSE._serialized_start=553
  _UPLOADDARFILERESPONSE._serialized_end=576
  _PACKAGEMANAGEMENTSERVICE._serialized_start=579
  _PACKAGEMANAGEMENTSERVICE._serialized_end=862
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Internal module used by dazl's build process to generate files from Protobuf/gRPC definitions.

**This is not a public API and its implementation details are subject to change at any time!**

The generated code was lightly inspired by https://github.com/dropbox/mypy-protobuf.
Its output differs in some key ways:

 * `mypy-protobuf` 's stubs are not (yet) compatible with the gRPC asyncio library.
 * Protobuf symbols (message names, enum values, service names, etc) are not allowed to start with
   underscores; as much as possible, we try to use underscore prefixes before all of our symbols
   (see `Protobuf Spec <https://developers.google.com/protocol-buffers/docs/reference/proto3-spec>`_).
"""



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.internal.containers import RepeatedScalarFieldContainer
from google.protobuf.message import Message as _Message
from google.protobuf.wrappers_pb2 import StringValue

from .value_pb2 import Identifier, Record, Value

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "Event",
    "CreatedEvent",
    "ArchivedEvent",
    "ExercisedEvent",
]


class Event(_Message):
    @property
    def created(self) -> CreatedEvent: ...
    @property
    def archived(self) -> ArchivedEvent: ...
    @_typing.overload
    def __init__(self): ...
    @_typing.overload
    def __init__(self, *, created: CreatedEvent = ...): ...
    @_typing.overload
    def __init__(self, *, archived: ArchivedEvent = ...): ...
    def HasField(self, field_name: _L["event", "created", "archived"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["event", "created", "archived"]) -> None: ...
    def WhichOneof(self, oneof_group: _L["event"]) -> _L[None, "created", "archived"]: ...

class CreatedEvent(_Message):
    event_id: _builtins.str
    contract_id: _builtins.str
    @property
    def template_id(self) -> Identifier: ...
    @property
    def contract_key(self) -> Value: ...
    @property
    def create_arguments(self) -> Record: ...
    @property
    def witness_parties(self) -> RepeatedScalarFieldContainer[_builtins.str]: ...
    @property
    def signatories(self) -> RepeatedScalarFieldContainer[_builtins.str]: ...
    @property
    def observers(self) -> RepeatedScalarFieldContainer[_builtins.str]: ...
    @property
    def agreement_text(self) -> StringValue: ...
    def __init__(self, *, event_id: _typing.Optional[_builtins.str] = ..., contract_id: _typing.Optional[_builtins.str] = ..., template_id: _typing.Optional[Identifier] = ..., contract_key: _typing.Optional[Value] = ..., create_arguments: _typing.Optional[Record] = ..., witness_parties: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., signatories: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., observers: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., agreement_text: _typing.Optional[StringValue] = ...): ...
    def HasField(self, field_name: _L["event_id", "contract_id", "template_id", "contract_key", "create_arguments", "witness_parties", "signatories", "observers", "agreement_text"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["event_id", "contract_id", "template_id", "contract_key", "create_arguments", "witness_parties", "signatories", "observers", "agreement_text"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ArchivedEvent(_Message):
    event_id: _builtins.str
    contract_id: _builtins.str
    @property
    def template_id(self) -> Identifier: ...
    @property
    def witness_parties(self) -> RepeatedScalarFieldContainer[_builtins.str]: ...
    def __init__(self, *, event_id: _typing.Optional[_builtins.str] = ..., contract_id: _typing.Optional[_builtins.str] = ..., template_id: _typing.Optional[Identifier] = ..., witness_parties: _typing.Optional[_typing.Iterable[_builtins.str]] = ...): ...
    def HasField(self, field_name: _L["event_id", "contract_id", "template_id", "witness_parties"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["event_id", "contract_id", "template_id", "witness_parties"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ExercisedEvent(_Message):
    event_id: _builtins.str
    contract_id: _builtins.str
    @property
    def template_id(self) -> Identifier: ...
    choice: _builtins.str
    @property
    def choice_argument(self) -> Value: ...
    @property
    def acting_parties(self) -> RepeatedScalarFieldContainer[_builtins.str]: ...
    consuming: _builtins.bool
    @property
    def witness_parties(self) -> RepeatedScalarFieldContainer[_builtins.str]: ...
    @property
    def child_event_ids(self) -> RepeatedScalarFieldContainer[_builtins.str]: ...
    @property
    def exercise_result(self) -> Value: ...
    def __init__(self, *, event_id: _typing.Optional[_builtins.str] = ..., contract_id: _typing.Optional[_builtins.str] = ..., template_id: _typing.Optional[Identifier] = ..., choice: _typing.Optional[_builtins.str] = ..., choice_argument: _typing.Optional[Value] = ..., acting_parties: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., consuming: _typing.Optional[_builtins.bool] = ..., witness_parties: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., child_event_ids: _typing.Optional[_typing.Iterable[_builtins.str]] = ..., exercise_result: _typing.Optional[Value] = ...): ...
    def HasField(self, field_name: _L["event_id", "contract_id", "template_id", "choice", "choice_argument", "acting_parties", "consuming", "witness_parties", "child_event_ids", "exercise_result"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["event_id", "contract_id", "template_id", "choice", "choice_argument", "acting_parties", "consuming", "witness_parties", "child_event_ids", "exercise_result"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from collections import defaultdict
from io import StringIO
import re
from typing import Collection, DefaultDict, Dict, Iterable, Mapping, Sequence, Set, Union

from google.protobuf.descriptor_pb2 import FieldDescriptorProto

from .symbols import SymbolTable, Usage
from .types import PyType

__all__ = ["ImportContext", "rewrite_file_content"]


FROM = re.compile(r"from ([\w.]+) import (\w+) as (\w+)")


class ImportContext:
    def __init__(self, parent: "SymbolTable"):
        self._imports = defaultdict(set)  # type: DefaultDict[str, Set[str]]
        self._system_imports = defaultdict(set)  # type: DefaultDict[str, Set[str]]
        self._parent = parent

    def add_system_import(self, __from: str, __import: str) -> None:
        self._system_imports[__from].add(__import)

    def add_import(self, __from: str, __import: str) -> None:
        """
        Add an import to this context.

        :param __from:
            The Python module to import _from_.
        :param __import:
            The Python module to import.
        """
        self._imports[__from].add(__import)

    def py_type(self, fd: "Union[FieldDescriptorProto, str]", usage: "Usage") -> "PyType":
        if isinstance(fd, str):
            fd = FieldDescriptorProto(type=FieldDescriptorProto.TYPE_MESSAGE, type_name=fd)

        pt = self._parent.py_type(fd, usage)
        for k, v in pt.imports.items():
            self._imports[k].update(v)
        return pt

    def required_imports(self) -> "Mapping[str, Sequence[str]]":
        return {__from: sorted(__import) for __from, __import in self._imports.items()}

    def py_import_block(self, relative_to: str) -> str:
        absolute_imports = {}  # type: Dict[str, Collection[str]]
        relative_imports = {}  # type: Dict[str, Collection[str]]

        for __from, __imports in sorted(self.required_imports().items()):
            if __from == relative_to:
                # we don't need to include imports to our own file
                continue

            elif __from.startswith("com.daml."):
                # rewrite our own imports as relative imports
                relative_imports[relative_package(relative_to, __from)] = __imports

            else:
                absolute_imports[__from] = __imports

        import_groups = [
            self._system_imports,
            absolute_imports,
            relative_imports,
        ]  # type: Iterable[Mapping[str, Collection[str]]]

        return "\n".join(
            render_import_group(import_group) for import_group in import_groups if import_group
        )


def relative_package(__current_package: str, __target_package: str) -> str:
    """
    Return a replacement string for the target package that can be used to import that target
    package, but in a way that is interpreted relative to the current package.

    :param __current_package:
    :param __target_package:
    :return:

    >>> relative_package("com.daml.ledger.api.v1.commands_pb2", "com.daml.ledger.api.v1.value_pb2")
    '.value_pb2'
    >>> relative_package("com.abc.def", "com.abc.xyz.ghi.jkl")
    '.xyz.ghi.jkl'
    >>> relative_package("com.abc.xyz.ghi.jkl", "com.abc.def")
    '...def'
    """
    cur_elems = __current_package.split(".")
    target_elems = __target_package.split(".")

    i = 0
    for i, (cur, target) in enumerate(zip(cur_elems, target_elems)):
        if cur != target:
            break

    # i is now the first index where the lists don't match
    return "." * (len(cur_elems) - i) + ".".join(target_elems[i:])


def render_import_group(import_group: "Mapping[str, Collection[str]]") -> str:
    with StringIO() as buf:
        for f, i in import_group.items():
            import_expr = ", ".join(sorted(i))
            if f:
                buf.write(f"from {f} import {import_expr}\n")
            else:
                buf.write(f"import {import_expr}\n")

        return buf.getvalue()


def rewrite_file_content(name: str, content: str) -> str:
    """
    Rewrite the content of generated Python field to use relative imports instead of absolute
    imports, with an assumed root module as implied by the name of the file.

    :param name: The name of the ``.proto`` file.
    :param content: The original content from the Protobuf compiler.
    :return: The contents of the file, with imports modified.
    """
    current_module = name.rpartition("/")[0].replace("/", ".")
    with StringIO() as out_buf, StringIO(content) as in_buf:
        for line in in_buf.readlines():
            out_buf.write(rewrite_import(current_module, line))
        return out_buf.getvalue()


def rewrite_import(parent_module: str, line: str) -> str:
    result = FROM.match(line)
    if result:
        module_name = result.group(1)
        identifier_name = result.group(2)
        local_name = result.group(3)
        if module_name.startswith("google."):
            # don't rewrite Google imports; simply do nothing
            return line
        if module_name == parent_module:
            return f"from . import {identifier_name} as {local_name}\n"

    return line



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

# TODO: `automodule reading` directive doesn't appear to work here, have to list each class individually.
"""
Read-Side Types
---------------

This module contains models used on the read-side of the Ledger API.

.. autoclass:: InitEvent
  :members:

.. autoclass::  InitEvent
  :members:

.. autoclass::  OffsetEvent
  :members:

.. autoclass::  ReadyEvent
  :members:

.. autoclass::  ActiveContractSetEvent
  :members:

.. autoclass::  BaseTransactionEvent
  :members:

.. autoclass::  TransactionStartEvent
  :members:

.. autoclass::  TransactionEndEvent
  :members:

.. autoclass::  ContractEvent
  :members:

.. autoclass::  ContractCreateEvent
  :members:

.. autoclass::  ContractExercisedEvent
  :members:

.. autoclass::  ContractArchiveEvent
  :members:

.. autoclass::  PackagesAddedEvent
  :members:

.. autoclass::  TransactionFilter
  :members:

.. autoclass::  EventKey
  :members:

"""

from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Any, Collection, Optional, Sequence, Union

from ..damlast.daml_lf_1 import TypeConName
from ..damlast.protocols import SymbolLookup
from ..prim import ContractData, ContractId, Party

if TYPE_CHECKING:
    from ..client.state import ContractContextualData
    from ..model.types_store import PackageStore

__all__ = [
    "BaseEvent",
    "InitEvent",
    "OffsetEvent",
    "ReadyEvent",
    "ActiveContractSetEvent",
    "BaseTransactionEvent",
    "TransactionStartEvent",
    "TransactionEndEvent",
    "ContractEvent",
    "ContractCreateEvent",
    "ContractExercisedEvent",
    "ContractArchiveEvent",
    "PackagesAddedEvent",
    "ContractFilter",
    "TransactionFilter",
]


@dataclass(frozen=True)
class BaseEvent:
    """
    Superclass of all dazl events.
    """

    client: "Any"
    party: "Optional[Party]"
    time: "Optional[datetime]"
    ledger_id: str
    lookup: "SymbolLookup"
    package_store: "PackageStore"

    def acs_find_active(self, template: "Union[str, TypeConName]", match=None):
        return self.client.find_active(template, match)

    def acs_find_by_id(self, cid: "Union[str, ContractId]") -> "Optional[ContractContextualData]":
        return self.client.find_by_id(cid)

    def acs_find_one(self, template: "Union[str, TypeConName]", match=None):
        return self.client.find_one(template, match=match)

    def acs_find_historical(self, template: "Union[str, TypeConName]", match=None):
        return self.client.find_historical(template, match)

    def acs_find_nonempty(self, template: "Union[str, TypeConName]", match=None):
        return self.client.find_nonempty(template, match=match)

    def __repr__(self):
        fields = ", ".join(
            f"{k}={v!r}"
            for k, v in self.__dict__.items()
            if not k.startswith("_") and k != "client" and k != "lookup" and k != "ledger_id"
        )
        return f"{self.__class__.__name__}({fields})"


@dataclass(frozen=True)
class InitEvent(BaseEvent):
    """
    Event raised when dazl is initialized, but before it has begun reading from the Active Contract
    Set (ACS).
    """


@dataclass(frozen=True)
class OffsetEvent(BaseEvent):
    """
    Event raised when dazl is ready to begin processing new events. At this point, the Active
    Contract Set (ACS) is populated with the current state of the ledger.
    """

    offset: Optional[str]


@dataclass(frozen=True)
class ReadyEvent(OffsetEvent):
    """
    Event raised when dazl is ready to begin processing new events. At this point, the Active
    Contract Set (ACS) is populated with the current state of the ledger.
    """


@dataclass(frozen=True)
class ActiveContractSetEvent(OffsetEvent):
    """
    Event raised on initial read of the active contract set.
    """

    contract_events: "Sequence[ContractCreateEvent]"


@dataclass(frozen=True)
class BaseTransactionEvent(OffsetEvent):
    """
    Event raised when dazl encounters a new transaction. This is raised before any corresponding
    :class:`ContractCreateEvent` or :class:`ContractArchiveEvent`.
    """

    command_id: str
    workflow_id: str


@dataclass(frozen=True)
class TransactionStartEvent(BaseTransactionEvent):
    """
    Event raised when dazl encounters a new transaction. This is raised before any corresponding
    :class:`ContractCreateEvent` or :class:`ContractArchiveEvent`.
    """

    contract_events: "Sequence[ContractEvent]"


@dataclass(frozen=True)
class TransactionEndEvent(BaseTransactionEvent):
    """
    Event raised when dazl encounters the end of a transaction. This is raised after any
    corresponding :class:`ContractCreateEvent` or :class:`ContractArchiveEvent`.
    """

    contract_events: "Sequence[ContractEvent]"


@dataclass(frozen=True)
class ContractEvent(BaseTransactionEvent):
    """
    Event raised when dazl automation detects a new create or an archive. The Active Contract Set
    (ACS) reflects this event, as well as all other events that occurred in the same transaction.
    """

    cid: ContractId
    cdata: ContractData
    command_id: str
    workflow_id: str
    event_id: str
    witness_parties: Sequence[str]


@dataclass(frozen=True)
class ContractCreateEvent(ContractEvent):
    """
    Event raised when dazl automation detects a contract create. The Active Contract Set
    (ACS) reflects this event, as well as all other events that occurred in the same transaction.
    """


@dataclass(frozen=True)
class ContractExercisedEvent(ContractEvent):
    """
    Event raised when dazl automation detects a contract exercised.
    """

    contract_creating_event_id: None
    choice: str
    choice_args: Any
    acting_parties: Sequence[str]
    consuming: bool
    child_event_ids: Sequence[str]
    exercise_result: Any


@dataclass(frozen=True)
class ContractArchiveEvent(ContractEvent):
    """
    Event raised when dazl automation detects a contract archive. The Active Contract Set
    (ACS) reflects this event, as well as all other events that occurred in the same transaction.
    """


@dataclass(frozen=True)
class PackagesAddedEvent(BaseEvent):
    """
    Event raised when new packages have been detected.
    """

    initial: bool


@dataclass(frozen=True)
class ContractFilter:
    templates: "Optional[Collection[TypeConName]]" = None
    party_groups: "Optional[Collection[Party]]" = None


@dataclass(frozen=True)
class TransactionFilter(ContractFilter):
    current_offset: "Optional[str]" = None
    destination_offset: "Optional[str]" = None

    def __post_init__(self):
        if self.current_offset is not None and self.destination_offset is not None:
            if self.current_offset > self.destination_offset:
                raise ValueError(
                    "current_offset must be before destination_offset if both are specified"
                )



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Pretty print format that orders templates by ID, and attempts to render a more compact
representation.
"""

from collections import defaultdict
from typing import AbstractSet, Generator, Iterable, Mapping, Sequence

from ... import LOG
from ...damlast.daml_lf_1 import Type, TypeConName
from ...damlast.daml_types import con
from ...damlast.protocols import SymbolLookup
from ...prim import Party
from ...values import ArrayStringMapper, Context
from .model import Formatter, RowBuilder

__all__ = ["PrettyFormatter"]

BOX_B3 = "|"
BOX_C4 = "-"
BOX_DA = "+"


class PrettyFormatter(Formatter):
    def render(
        self, lookup: "SymbolLookup", parties: "AbstractSet[Party]", entries: "Iterable[RowBuilder]"
    ):
        sort = ByPartySort(parties)

        col_spacer = "  "

        entries_by_template = group_by_name(entries)
        entry_count = sum(len(entries) for entries in entries_by_template.values())

        yield "{} total contracts over {} templates".format(entry_count, len(entries_by_template))
        for line in party_header(parties):
            yield line

        for name, entries in entries_by_template.items():
            # for each column for all entries in the data set, determine the most
            # compact representation here
            yield ""

            tt = con(name)
            context = Context(ArrayStringMapper(), lookup)
            header_row = ["", "#cid", *(expand_record_field_names(context, Type.Con(name, ())))]

            with LOG.debug_timed("Render entries as strings"):
                orig_rows = [
                    (
                        sort.key(entry),
                        [
                            render_parties(parties, entry),
                            str(entry.cid),
                            *context.convert(tt, entry.cdata),
                        ],
                    )
                    for entry in entries
                ]

            with LOG.debug_timed("Sort rows"):
                orig_rows.sort(key=lambda val: val[0])
                rows = [row for _, row in orig_rows]
                rows.insert(0, header_row)

            with LOG.debug_timed("Measure entry sizes"):
                cell_sizes = [0] * len(rows[0])
                for row in rows:
                    for i, cell in enumerate(row):
                        cell_size = len(cell)
                        if cell_sizes[i] < cell_size:
                            cell_sizes[i] = cell_size

            with LOG.debug_timed("Render rows"):
                header_len = sum(cell_sizes) + (len(cell_sizes) - 1) * len(col_spacer)
                header = f"{name} ({contract_count(len(entries))}) "

                yield header + "-" * max(3, header_len - len(header))
                for row in rows:
                    yield col_spacer.join(
                        cell.ljust(cell_size) for cell_size, cell in zip(cell_sizes, row)
                    )


def contract_count(c: int) -> str:
    if c == 1:
        return "1 contract"
    else:
        return f"{c} contracts"


def expand_record_field_names(
    context: "Context", con: "Type.Con", path: "Sequence[str]" = ()
) -> "Generator[str, None, None]":
    """
    From a starting :class:`Type.Con`, return a string list of field names, additionally expanding
    sub-record fields.

    :param context:
        A context to use to perform type evaluation.
    :param con:
    :param path:
    :return:
    """
    dt = context.resolve_data_type(con)
    if dt.record is not None:
        for field in dt.record.fields:
            subpath = (*path, field.field)
            if field.type.con is not None:
                child_fields = list(expand_record_field_names(context, field.type.con, subpath))
                if child_fields:
                    yield from child_fields
                else:
                    yield ".".join(subpath)
            else:
                yield ".".join(subpath)


def party_header(parties):
    parties = tuple(parties)
    party_count = len(parties)

    for index, party in enumerate(parties):
        ascii_art = (BOX_B3 * index) + BOX_DA + (BOX_C4 * (party_count - index - 1))
        yield f"{ascii_art} party '{party}'"
    yield "|" * party_count


class ByPartySort:
    def __init__(self, parties):
        self.parties = parties

    def key(self, entry):
        party_vis = [1 if entry.parties.get(party) is not None else 0 for party in self.parties]
        return sum(party_vis), "".join(map(str, reversed(party_vis)))


def render_parties(all_parties, entry):
    return "".join(_render_party_bool(entry.parties.get(party)) for party in all_parties)


def _render_party_bool(value):
    if value is not None:
        return "C" if value else "A"
    return " "


def group_by_name(entries: "Iterable[RowBuilder]") -> "Mapping[TypeConName, Sequence[RowBuilder]]":
    """
    Organize the entries by their template name. The returned mapping where the keys are in sorted
    order.
    """
    entries_by_template = defaultdict(list)
    for entry in entries:
        entries_by_template[entry.cid.value_type].append(entry)

    return {name: entries_by_template[name] for name in sorted(entries_by_template)}



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# type: ignore

from io import StringIO
from typing import Dict, List, Tuple

from dazl.model.types import ContractIdType, ListType, RecordType, Type, TypeReference
from dazl.model.types_store import PackageStore


def write_csv(store: PackageStore) -> str:
    templates = {
        template.data_type.name.full_name: template for template in store.resolve_template("*")
    }

    with StringIO() as buf:
        for template_name in sorted(templates):
            template = templates[template_name]
            buf.write(f"# {template_name}\n")

            field_types = {}  # type: Dict[str, Type]
            queue = [("", template.data_type)]  # type: List[Tuple[str, Type]]
            while queue:
                prefix, data_type = queue.pop(0)
                if isinstance(data_type, TypeReference):
                    data_type = store.resolve_type_reference(data_type)
                if isinstance(data_type, RecordType):
                    pfx = (prefix + ".") if prefix else ""
                    queue[0:0] = [(f"{pfx}{n}", dt) for n, dt in data_type.named_args]
                elif isinstance(data_type, ListType):
                    queue.insert(0, (f"{prefix}[]", data_type.type_parameter))
                else:
                    field_types[prefix] = data_type

            for name, terminal_type in field_types.items():
                if isinstance(terminal_type, ContractIdType):
                    buf.write(f"#    {name}: ContractId {terminal_type.type_parameter}\n")
                else:
                    buf.write(f"#    {name}: {terminal_type}\n")
            buf.write(",".join(f'"{k}"' for k in field_types))
            buf.write("\n\n")

        return buf.getvalue()



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from decimal import Decimal
from typing import Any, Optional

__all__ = ["to_int", "to_decimal", "decimal_to_str"]


def to_int(obj: "Any") -> int:
    if isinstance(obj, int):
        return obj
    elif isinstance(obj, str):
        return int(obj)
    elif isinstance(obj, Decimal):
        return int(obj)
    raise ValueError(f"Could not parse as an int: {obj!r}")


def to_decimal(obj: "Optional[Any]") -> "Optional[Decimal]":
    """
    Convert any of the common wire representations of a ``Decimal`` to a ``Decimal``.
    """
    if obj is None:
        # noinspection PyTypeChecker
        return None

    if isinstance(obj, Decimal):
        return obj
    elif isinstance(obj, (int, float)):
        return Decimal(str(obj))
    elif isinstance(obj, str):
        return Decimal(obj)
    raise ValueError(f"Could not parse as a Decimal: {obj!r}")


def decimal_to_str(d: "Decimal") -> str:
    """
    Return a string representation of a :class:`Decimal` that is safe to be used over the
    Ledger API. Concretely, this means staying away from scientific notation.
    """
    precision = max(0, -d.as_tuple().exponent)
    return format(d, f".{precision}f")



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Logging utilities used in `dazl`.

This is an internal API and not meant to be used directly outside of dazl; symbols declared in this
file may change at any time.
"""

from functools import partial
import logging
import sys
from time import time
from typing import Type, cast

__all__ = ["LOG", "VERBOSE", "configure"]

# A custom logging level below "DEBUG". This is used within dazl for messages that are rarely
# important enough to be printed, even within tests, but still occasionally have value when
# debugging particular blocks of code.
VERBOSE = 5
logging.addLevelName(VERBOSE, "VERBOSE")

# If a custom logger type was set before dazl was loaded, respect it by using it as our logger's
# base class.
# noinspection PyTypeChecker
Logger = logging.getLoggerClass()  # type: Type[logging.Logger]


class ExtendedLogger(Logger):  # type: ignore
    """
    A logger with additional utility logging functions used within dazl.
    """

    def verbose(self, msg, *args, **kwargs):
        """
        Log a message with level ``VERBOSE`` on this logger.
        """
        self.log(VERBOSE, msg, *args, **kwargs)

    def verbose_timed(self, msg, *args, **kwargs) -> "TimedLogMessageContext":
        """
        Log a message with level ``VERBOSE`` on the logger, additionally annotating the log message
        with the time it took to complete the block.
        """
        return TimedLogMessageContext(self, VERBOSE, msg, args, kwargs)

    def debug_timed(self, msg, *args, **kwargs) -> "TimedLogMessageContext":
        """
        Log a message with level ``DEBUG`` on the logger, additionally annotating the log message
        with the time it took to complete the block.
        """
        return TimedLogMessageContext(self, logging.DEBUG, msg, args, kwargs)

    def info_timed(self, msg, *args, **kwargs) -> "TimedLogMessageContext":
        """
        Log a message with level ``INFO`` on the logger, additionally annotating the log message
        with the time it took to complete the block.
        """
        return TimedLogMessageContext(self, logging.INFO, msg, args, kwargs)


# Create our Logger with our special type that has all of our goodies.
logging.setLoggerClass(ExtendedLogger)
LOG = cast(ExtendedLogger, logging.getLogger("dazl"))

# There is a chance that someone instantiated a logger using our name already. If that's the case,
# then it is too late to specify the actual class instance used and `LOG` will not have the
# functions declared on it that we expect. So make sure that we specifically add methods to _our_
# logger.
#
# An alternate implementation would have been to simply assign our custom logging functions directly
# to an instance logger without playing games with `logger.setLoggerClass`; but we don't, because
# mypy/IDEs won't understand what is going on; we'd be using "reflection" instead of giving
# mypy/IDEs a simple class (ExtendedLogger) to key off of.
for field in dir(ExtendedLogger):
    if not field.startswith("__") and not hasattr(LOG, field):
        setattr(LOG, field, partial(getattr(ExtendedLogger, field), LOG))

# This is essentially a runtime assertion that our logging functions have been defined properly.
LOG.verbose("dazl logging has been loaded.")

# Restore the original Logger type.
logging.setLoggerClass(Logger)


class TimedLogMessageContext:
    """
    A simple ContextManager that logs a message at the specified logging level, additionally with
    timing information. If an exception is thrown in the block, that exception is instead logged.
    """

    __slots__ = "logger", "log_level", "msg", "args", "kwargs", "start"

    def __init__(self, logger, log_level, msg, args, kwargs):
        self.logger = logger
        self.log_level = log_level
        self.msg = msg
        self.args = args
        self.kwargs = kwargs
        self.start = None

    def __enter__(self):
        self.start = time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        elapsed_ms = (time() - self.start) * 1000.0
        if exc_type is not None:
            self.logger.exception(
                self.msg + " (%0.2f ms)",
                *self.args,
                elapsed_ms,
                exc_info=(exc_type, exc_val, exc_tb),
                **self.kwargs
            )
        else:
            self.logger.log(
                self.log_level, self.msg + " (%0.2f ms)", *self.args, elapsed_ms, **self.kwargs
            )


did_configure = False


def configure(level=logging.INFO):
    """
    Set up a default logger format and stream handler with sensible defaults.
    """
    root = logging.getLogger()
    global did_configure
    if not did_configure:
        did_configure = True
    else:
        root.warning("configure being called more than once!")
        return

    logging.captureWarnings(True)

    root.setLevel(level)
    stream_handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter("[%(levelname)7s] %(asctime)s | %(name)-7s | %(message)s")
    stream_handler.setFormatter(formatter)
    root.addHandler(stream_handler)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
from google.protobuf.message import Message as _Message
from google.protobuf.timestamp_pb2 import Timestamp

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "GetMeteringReportRequest",
    "GetMeteringReportResponse",
    "ParticipantMeteringReport",
    "ApplicationMeteringReport",
]


class GetMeteringReportRequest(_Message):
    @property
    def to(self) -> Timestamp: ...
    application_id: _builtins.str
    def __init__(self, *, to: _typing.Optional[Timestamp] = ..., application_id: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["from", "to", "application_id"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["from", "to", "application_id"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class GetMeteringReportResponse(_Message):
    @property
    def request(self) -> GetMeteringReportRequest: ...
    @property
    def participant_report(self) -> ParticipantMeteringReport: ...
    @property
    def report_generation_time(self) -> Timestamp: ...
    def __init__(self, *, request: _typing.Optional[GetMeteringReportRequest] = ..., participant_report: _typing.Optional[ParticipantMeteringReport] = ..., report_generation_time: _typing.Optional[Timestamp] = ...): ...
    def HasField(self, field_name: _L["request", "participant_report", "report_generation_time"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["request", "participant_report", "report_generation_time"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ParticipantMeteringReport(_Message):
    participant_id: _builtins.str
    is_final: _builtins.bool
    @property
    def application_reports(self) -> RepeatedCompositeFieldContainer[ApplicationMeteringReport]: ...
    def __init__(self, *, participant_id: _typing.Optional[_builtins.str] = ..., is_final: _typing.Optional[_builtins.bool] = ..., application_reports: _typing.Optional[_typing.Iterable[ApplicationMeteringReport]] = ...): ...
    def HasField(self, field_name: _L["participant_id", "is_final", "application_reports"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["participant_id", "is_final", "application_reports"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ApplicationMeteringReport(_Message):
    application_id: _builtins.str
    event_count: _builtins.int
    def __init__(self, *, application_id: _typing.Optional[_builtins.str] = ..., event_count: _typing.Optional[_builtins.int] = ...): ...
    def HasField(self, field_name: _L["application_id", "event_count"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["application_id", "event_count"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...
